# From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
# and IDIAP Research Institute nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# This file is copied from https://github.com/pytorch/pytorch/tree/v1.9.0/torch/testing/_internal/common_cuda.py
r"""This file is allowed to initialize CUDA context when imported."""

import functools
import torch
import torch.cuda
from tests.internal.torch.common_utils import TEST_NUMBA
import inspect
import contextlib
from setuptools import distutils


TEST_CUDA = torch.cuda.is_available()
TEST_MULTIGPU = TEST_CUDA and torch.cuda.device_count() >= 2
CUDA_DEVICE = torch.device("cuda:0") if TEST_CUDA else None
# note: if ROCm is targeted, TEST_CUDNN is code for TEST_MIOPEN
TEST_CUDNN = TEST_CUDA and torch.backends.cudnn.is_acceptable(
    torch.tensor(1.0, device=CUDA_DEVICE)
)
TEST_CUDNN_VERSION = torch.backends.cudnn.version() if TEST_CUDNN else 0

CUDA11OrLater = (
    torch.version.cuda
    and distutils.version.LooseVersion(torch.version.cuda) >= "11.0.0"
)
CUDA9 = torch.version.cuda and torch.version.cuda.startswith("9.")
SM53OrLater = torch.cuda.is_available() and torch.cuda.get_device_capability() >= (5, 3)

TEST_MAGMA = TEST_CUDA
if TEST_CUDA:
    torch.ones(1).cuda()  # has_magma shows up after cuda is initialized
    TEST_MAGMA = torch.cuda.has_magma

if TEST_NUMBA:
    import numba.cuda

    TEST_NUMBA_CUDA = numba.cuda.is_available()
else:
    TEST_NUMBA_CUDA = False

# Used below in `initialize_cuda_context_rng` to ensure that CUDA context and
# RNG have been initialized.
__cuda_ctx_rng_initialized = False


# after this call, CUDA context and RNG must have been initialized on each GPU
def initialize_cuda_context_rng():
    global __cuda_ctx_rng_initialized
    assert TEST_CUDA, "CUDA must be available when calling initialize_cuda_context_rng"
    if not __cuda_ctx_rng_initialized:
        # initialize cuda context and rng for memory tests
        for i in range(torch.cuda.device_count()):
            torch.randn(1, device="cuda:{}".format(i))
        __cuda_ctx_rng_initialized = True


# Test whether hardware TF32 math mode enabled. It is enabled only on:
# - CUDA >= 11
# - arch >= Ampere
def tf32_is_not_fp32():
    if not torch.cuda.is_available() or torch.version.cuda is None:
        return False
    if torch.cuda.get_device_properties(torch.cuda.current_device()).major < 8:
        return False
    if int(torch.version.cuda.split(".")[0]) < 11:
        return False
    return True


@contextlib.contextmanager
def tf32_off():
    old_allow_tf32_matmul = torch.backends.cuda.matmul.allow_tf32
    try:
        torch.backends.cuda.matmul.allow_tf32 = False
        with torch.backends.cudnn.flags(
            enabled=None, benchmark=None, deterministic=None, allow_tf32=False
        ):
            yield
    finally:
        torch.backends.cuda.matmul.allow_tf32 = old_allow_tf32_matmul


@contextlib.contextmanager
def tf32_on(self, tf32_precision=1e-5):
    old_allow_tf32_matmul = torch.backends.cuda.matmul.allow_tf32
    old_precison = self.precision
    try:
        torch.backends.cuda.matmul.allow_tf32 = True
        self.precision = tf32_precision
        with torch.backends.cudnn.flags(
            enabled=None, benchmark=None, deterministic=None, allow_tf32=True
        ):
            yield
    finally:
        torch.backends.cuda.matmul.allow_tf32 = old_allow_tf32_matmul
        self.precision = old_precison


# This is a wrapper that wraps a test to run this test twice, one with
# allow_tf32=True, another with allow_tf32=False. When running with
# allow_tf32=True, it will use reduced precision as pecified by the
# argument. For example:
#    @dtypes(torch.float32, torch.float64, torch.complex64, torch.complex128)
#    @tf32_on_and_off(0.005)
#    def test_matmul(self, device, dtype):
#        a = ...; b = ...;
#        c = torch.matmul(a, b)
#        self.assertEqual(c, expected)
# In the above example, when testing torch.float32 and torch.complex64 on CUDA
# on a CUDA >= 11 build on an >=Ampere architecture, the matmul will be running at
# TF32 mode and TF32 mode off, and on TF32 mode, the assertEqual will use reduced
# precision to check values.
#
# This decorator can be used for function with or without device/dtype, such as
# @tf32_on_and_off(0.005)
# def test_my_op(self)
# @tf32_on_and_off(0.005)
# def test_my_op(self, device)
# @tf32_on_and_off(0.005)
# def test_my_op(self, device, dtype)
# @tf32_on_and_off(0.005)
# def test_my_op(self, dtype)
# if neither device nor dtype is specified, it will check if the system has ampere device
# if device is specified, it will check if device is cuda
# if dtype is specified, it will check if dtype is float32 or complex64
# tf32 and fp32 are different only when all the three checks pass
def tf32_on_and_off(tf32_precision=1e-5):
    def with_tf32_disabled(self, function_call):
        with tf32_off():
            function_call()

    def with_tf32_enabled(self, function_call):
        with tf32_on(self, tf32_precision):
            function_call()

    def wrapper(f):
        params = inspect.signature(f).parameters
        arg_names = tuple(params.keys())

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            for k, v in zip(arg_names, args):
                kwargs[k] = v
            cond = tf32_is_not_fp32()
            if "device" in kwargs:
                cond = cond and (torch.device(kwargs["device"]).type == "cuda")
            if "dtype" in kwargs:
                cond = cond and (kwargs["dtype"] in {torch.float32, torch.complex64})
            if cond:
                with_tf32_disabled(kwargs["self"], lambda: f(**kwargs))
                with_tf32_enabled(kwargs["self"], lambda: f(**kwargs))
            else:
                f(**kwargs)

        return wrapped

    return wrapper


# This is a wrapper that wraps a test to run it with TF32 turned off.
# This wrapper is designed to be used when a test uses matmul or convolutions
# but the purpose of that test is not testing matmul or convolutions.
# Disabling TF32 will enforce torch.float tensors to be always computed
# at full precision.
def with_tf32_off(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        with tf32_off():
            return f(*args, **kwargs)

    return wrapped


def _get_torch_cuda_version():
    if torch.version.cuda is None:
        return (0, 0)
    cuda_version = str(torch.version.cuda)
    return tuple(int(x) for x in cuda_version.split("."))



import copy
import requests
import os
import time
import threading
import json
import logging
import multiprocessing
from .autotune_task_manager import AutotuneTaskManager
from bagua.bagua_define import (
    TensorDtype,
    TensorDeclaration,
    BaguaCoreTelemetrySpan,
    BaguaHyperparameter,
)
from flask import request
import numpy as np
from typing import Dict, List


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, TensorDtype):
            return obj.value
        else:
            return super(NpEncoder, self).default(obj)


class AutotuneServiceTaskManager:
    def __init__(
        self, task_name: str, world_size: int, is_output_autotune_log: bool
    ) -> None:
        self.inner = AutotuneTaskManager(task_name, is_output_autotune_log)
        self.warmup_pass_count = 0
        self.sampling_count = 0
        self.lock = threading.Lock()
        self.check_board = [-1] * world_size
        self.time_hp_last_granted = time.time()
        self.hyperparameter = BaguaHyperparameter()


class AutotuneService:
    MAX_TRACE_INFO = 1000

    def __init__(
        self,
        world_size,
        autotune_level=0,
        max_samples=60,
        sampling_confidence_time_s=5,
        warmup_time_s=30,
        is_output_autotune_log=False,
        default_bucket_size=10 * 1024 ** 2,
    ):
        self.autotune_level = autotune_level
        self.world_size = world_size
        self.max_samples = max_samples
        self.sampling_confidence_time_s = sampling_confidence_time_s
        self.warmup_time_s = warmup_time_s
        self.is_initialized = False
        self.is_output_autotune_log = is_output_autotune_log
        self.default_bucket_size: int = default_bucket_size
        self.model_dict: Dict[str, AutotuneServiceTaskManager] = {}
        self.model_dict_mutex = threading.Lock()

        # bagua-core trace and obtain tensor calculation partial order
        self.trace_info_dict = {}
        self.tensor_partial_order = {}
        self.tensor_partial_order_fixed = False
        self.tensor_partial_order_lock = threading.Lock()

    def autotune(
        self,
        hp_manager: AutotuneServiceTaskManager,
        rank: int,
        train_iter: int,
        tensor_partial_order: Dict[str, int] = {},
    ):
        if hp_manager.sampling_count > self.max_samples:
            return

        (
            recommended_train_iter,
            hp,
            system_efficiency_score,
        ) = hp_manager.inner.tail_record()

        logging.info(
            "recommended_train_iter={}, hyperparameters={}, speed={}".format(
                recommended_train_iter,
                hp,
                system_efficiency_score,
            )
        )

        sampling_time = time.time() - hp_manager.time_hp_last_granted

        # Skip at least once during warmup
        if (
            sampling_time < self.warmup_time_s
            or hp_manager.warmup_pass_count == 0  # noqa: W503
        ):
            logging.info(
                "warmup pass, time.time={}, time_hp_last_granted={}, "
                "warmup_time_s={}".format(
                    time.time(),
                    hp_manager.time_hp_last_granted,
                    self.warmup_time_s,
                )
            )
            hp_manager.warmup_pass_count += 1
            return

        if hp_manager.sampling_count == 0:
            confidence_skip = (
                sampling_time < self.warmup_time_s + self.sampling_confidence_time_s
            )
        else:
            confidence_skip = sampling_time < self.sampling_confidence_time_s

        if confidence_skip:
            logging.debug(
                "The sampling time is not up, time={}, last={}, "
                "sampling_confidence_time_s={}".format(
                    time.time(),
                    hp_manager.time_hp_last_granted,
                    self.sampling_confidence_time_s,
                )
            )
            return

        logging.info(
            "rank={}, train_iter={}, sampling_count={}, "
            "max_samples={}".format(
                rank, train_iter, hp_manager.sampling_count, self.max_samples
            )
        )
        recommended_bagua_hp = hp_manager.inner.ask_hyperparmeter(
            train_iter, tensor_partial_order
        )
        if hp_manager.sampling_count < self.max_samples:
            hp_manager.hyperparameter = recommended_bagua_hp
        else:
            hp_manager.hyperparameter = hp_manager.inner.best_hyperparameter()

        hp_manager.sampling_count += 1

    def setup_app(self, app):
        @app.route("/api/v1/register_tensors", methods=["POST"])
        def register_tensors():
            req: dict = request.get_json(force=True)
            model_name: str = req["model_name"]
            tensor_list: List[TensorDeclaration] = req["tensor_list"]
            whether_to_bucket: bool = req["whether_to_bucket"]

            with self.model_dict_mutex:
                if model_name not in self.model_dict:
                    self.model_dict[model_name] = AutotuneServiceTaskManager(
                        task_name=model_name,
                        world_size=self.world_size,
                        is_output_autotune_log=self.is_output_autotune_log,
                    )

            hp_manager = self.model_dict[model_name]
            bucket_size = self.default_bucket_size
            if whether_to_bucket is False:
                bucket_size = 10 * 1024 ** 5

            with hp_manager.lock:
                hp = BaguaHyperparameter(
                    buckets=AutotuneTaskManager.split_bucket_by_bucket_size(
                        tensor_list,
                        bucket_size,
                    ),
                    bucket_size=bucket_size,
                )
                hp_manager.time_hp_last_granted = time.time()
                hp_manager.hyperparameter = hp
                return json.dumps(
                    {
                        "recommended_hyperparameters": hp.dict(),
                    }
                )

        @app.route("/api/v1/report_metrics", methods=["POST"])
        def report_metrics():
            req: dict = request.get_json(force=True)
            model_name: str = req["model_name"]
            rank: int = req["rank"]
            train_iter: int = req["train_iter"]
            speed: float = req["speed"]
            hyperparameters = req["hyperparameters"]

            if model_name not in self.model_dict:
                return "Service not ready for report_metrics!", 405

            hp_manager = self.model_dict[model_name]

            # Only consider the rank of the first report metrics now.
            with hp_manager.lock:
                (last_report_train_iter, _, _) = hp_manager.inner.tail_record()
                if train_iter <= last_report_train_iter:
                    return json.dumps({})

                logging.debug(
                    "rank={}, train_iter={}, speed={}, "
                    "hyperparameters={}".format(
                        rank,
                        train_iter,
                        speed,
                        hyperparameters,
                    )
                )
                hp_manager.inner.report_metrics(
                    train_iter=train_iter,
                    hyperparameter=BaguaHyperparameter().update(hyperparameters),
                    system_efficiency_score=speed,
                )

            return json.dumps({})

        @app.route("/api/v1/ask_hyperparameters", methods=["POST"])
        def ask_hyperparameters():
            """
            report_metrics must be called before ask_hyperparameters
            """
            req: dict = request.get_json(force=True)
            rank: int = req["rank"]
            model_name: str = req["model_name"]
            train_iter: int = req["train_iter"]

            if model_name not in self.model_dict:
                return "Service not ready for report_metrics!", 405

            hp_manager = self.model_dict[model_name]

            tensor_partial_order = {}
            with self.tensor_partial_order_lock:
                tensor_partial_order = copy.deepcopy(self.tensor_partial_order)

            logging.debug("tensor_partial_order={}".format(tensor_partial_order))

            with hp_manager.lock:
                # Autotune conditions:
                # 1. autotune_level >= 1.
                # 2. The bagua process is not in the process of hyperparameter update. (self.check_board.count(self.check_board[0])
                #   == len(self.check_board))
                # 3. Only execute autotune at most once in an iteration. (self.check_board[rank] < train_iter)
                check_board = hp_manager.check_board
                if (
                    self.autotune_level >= 1
                    and check_board.count(check_board[0])  # noqa: W503
                    == len(check_board)  # noqa: W503
                    and check_board[rank] < train_iter  # noqa: W503
                ):
                    self.autotune(hp_manager, rank, train_iter, tensor_partial_order)

                check_board[rank] = train_iter

                return json.dumps(
                    {
                        "recommended_hyperparameters": hp_manager.hyperparameter.dict(),
                        "is_autotune_completed": hp_manager.sampling_count
                        > self.max_samples,  # noqa: W503
                    }
                )

        @app.route("/api/v1/report_tensor_execution_order", methods=["POST"])
        def report_tensor_execution_order():
            req: dict = request.get_json(force=True)
            spans: List[BaguaCoreTelemetrySpan] = req["spans"]

            with self.tensor_partial_order_lock:
                spans = sorted(spans, key=lambda span: span["start_time"])
                for span in spans:
                    tensor_name = span["tensor_name"]
                    action = span["action"]

                    if (tensor_name, action) in self.trace_info_dict:
                        continue

                    self.trace_info_dict[(tensor_name, action)] = True
                    if tensor_name not in self.tensor_partial_order:
                        self.tensor_partial_order[tensor_name] = len(
                            self.tensor_partial_order
                        )

            return json.dumps({})

        @app.route("/api/v1/health_check", methods=["GET"])
        def health_check():
            return json.dumps({"status": "ok"})

        # set secret-key
        app.config.update(SECRET_KEY=os.urandom(24))

        return app


def reset_error_retry(request_func):
    """Retry request when catch ConnectionResetError."""

    def wrap(*args, **kwargs):
        MAX_RETRIES = 3

        for retry in range(MAX_RETRIES + 1):
            try:
                result = request_func(*args, **kwargs)
                return result
            except (ConnectionResetError, requests.exceptions.ConnectionError) as e:
                if retry == MAX_RETRIES:
                    raise e
                logging.warning("request failed, retry={}, e={}".format(retry, e))
                time.sleep(1)

    return wrap


class AutotuneClient:
    def __init__(
        self,
        service_addr: str,
        service_port: int,
        proxies={
            "http": None,
            "https": None,
        },
    ):
        self.autotune_service_addr = "{}:{}".format(service_addr, service_port)
        self.session = requests.Session()
        self.proxies = proxies

        import socket
        from urllib3.connection import HTTPConnection

        HTTPConnection.default_socket_options = HTTPConnection.default_socket_options + [
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),  # Enables the feature
            (socket.SOL_TCP, socket.TCP_KEEPIDLE, 45),  # Overrides the time when the stack willl start sending KeppAlives after no data received on a Persistent Connection
            (socket.SOL_TCP, socket.TCP_KEEPINTVL, 10),  # Defines how often thoe KA will be sent between them
            (socket.SOL_TCP, socket.TCP_KEEPCNT, 6),  # How many attemps will your code try if the server goes down before droping the connection.
        ]

    @reset_error_retry
    def report_metrics(
        self,
        model_name: str,
        rank: int,
        train_iter: int,
        hyperparameters: dict,
        speed: float,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/report_metrics".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "rank": rank,
                "train_iter": train_iter,
                "hyperparameters": hyperparameters,
                "speed": speed,
            },
            proxies=self.proxies,
        )
        return rsp

    @reset_error_retry
    def register_tensors(
        self,
        model_name: str,
        tensor_list: List[TensorDeclaration],
        whether_to_bucket: bool = True,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/register_tensors".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "tensor_list": tensor_list,
                "whether_to_bucket": whether_to_bucket,
            },
            proxies=self.proxies,
        )
        return rsp

    @reset_error_retry
    def ask_hyperparameters(
        self,
        model_name: str,
        rank: int,
        train_iter: int,
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/ask_hyperparameters".format(self.autotune_service_addr),
            json={
                "model_name": model_name,
                "rank": rank,
                "train_iter": train_iter,
            },
            proxies=self.proxies,
        )
        return rsp

    @reset_error_retry
    def report_tensor_execution_order(
        self,
        spans: List[BaguaCoreTelemetrySpan],
    ) -> requests.Response:
        rsp = self.session.post(
            "http://{}/api/v1/report_tensor_execution_order".format(
                self.autotune_service_addr
            ),
            json={
                "spans": spans,
            },
            proxies=self.proxies,
        )
        return rsp

    def health_check(self) -> bool:
        try:
            # get response will be ok
            self.session.get(
                "http://{}/api/v1/health_check".format(
                    self.autotune_service_addr
                ),
                proxies=self.proxies
            )

            return True
        except requests.exceptions.ConnectionError:
            return False

if __name__ == "__main__":
    import argparse
    from flask import Flask

    parser = argparse.ArgumentParser()
    parser.add_argument("--nprocs", type=int, default=8)
    parser.add_argument("--port", type=int, default=8123)

    args = parser.parse_args()

    autotune_service = AutotuneService(args.nproc)
    app = Flask(__name__)
    app = autotune_service.setup_app(app)

    server = multiprocessing.Process(
        target=app.run,
        kwargs={
            "host": "0.0.0.0",
            "port": args.port,
        },
    )
    server.daemon = True
    server.start()
    server.join()



# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from torch.utils.data import DataLoader, Dataset, IterableDataset, Subset

from pytorch_lightning import LightningDataModule, LightningModule


class RandomDictDataset(Dataset):
    def __init__(self, size: int, length: int):
        self.len = length
        self.data = torch.randn(length, size)

    def __getitem__(self, index):
        a = self.data[index]
        b = a + 2
        return {"a": a, "b": b}

    def __len__(self):
        return self.len


class RandomDataset(Dataset):
    def __init__(self, size: int, length: int):
        self.len = length
        self.data = torch.randn(length, size)

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len


class RandomIterableDataset(IterableDataset):
    def __init__(self, size: int, count: int):
        self.count = count
        self.size = size

    def __iter__(self):
        for _ in range(self.count):
            yield torch.randn(self.size)


class RandomIterableDatasetWithLen(IterableDataset):
    def __init__(self, size: int, count: int):
        self.count = count
        self.size = size

    def __iter__(self):
        for _ in range(len(self)):
            yield torch.randn(self.size)

    def __len__(self):
        return self.count


class BoringModel(LightningModule):
    def __init__(self):
        """Testing PL Module.

        Use as follows:
        - subclass
        - modify the behavior for what you want

        class TestModel(BaseTestModel):
                def training_step(...):
                        # do your own thing

        or:

        model = BaseTestModel()
        model.training_epoch_end = None
        """
        super().__init__()
        self.layer = torch.nn.Linear(32, 2)

    def forward(self, x):
        return self.layer(x)

    def loss(self, batch, prediction):
        # An arbitrary loss to have a loss that updates the model weights during `Trainer.fit` calls
        return torch.nn.functional.mse_loss(prediction, torch.ones_like(prediction))

    def step(self, x):
        x = self(x)
        out = torch.nn.functional.mse_loss(x, torch.ones_like(x))
        return out

    def training_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        return {"loss": loss}

    def training_step_end(self, training_step_outputs):
        return training_step_outputs

    def training_epoch_end(self, outputs) -> None:
        torch.stack([x["loss"] for x in outputs]).mean()

    def validation_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        return {"x": loss}

    def validation_epoch_end(self, outputs) -> None:
        torch.stack([x["x"] for x in outputs]).mean()

    def test_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        return {"y": loss}

    def test_epoch_end(self, outputs) -> None:
        torch.stack([x["y"] for x in outputs]).mean()

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.layer.parameters(), lr=0.1)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1)
        return [optimizer], [lr_scheduler]

    def train_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def val_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def test_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def predict_dataloader(self):
        return DataLoader(RandomDataset(32, 64))


class BoringDataModule(LightningDataModule):
    def __init__(self, data_dir: str = "./"):
        super().__init__()
        self.data_dir = data_dir
        self.non_picklable = None
        self.checkpoint_state: Optional[str] = None
        self.random_full = RandomDataset(32, 64 * 4)

    def setup(self, stage: Optional[str] = None):
        if stage == "fit" or stage is None:
            self.random_train = Subset(self.random_full, indices=range(64))

        if stage in ("fit", "validate") or stage is None:
            self.random_val = Subset(self.random_full, indices=range(64, 64 * 2))

        if stage == "test" or stage is None:
            self.random_test = Subset(self.random_full, indices=range(64 * 2, 64 * 3))

        if stage == "predict" or stage is None:
            self.random_predict = Subset(
                self.random_full, indices=range(64 * 3, 64 * 4)
            )

    def train_dataloader(self):
        return DataLoader(self.random_train)

    def val_dataloader(self):
        return DataLoader(self.random_val)

    def test_dataloader(self):
        return DataLoader(self.random_test)

    def predict_dataloader(self):
        return DataLoader(self.random_predict)


class ManualOptimBoringModel(BoringModel):
    def __init__(self):
        super().__init__()
        self.automatic_optimization = False

    def training_step(self, batch, batch_idx):
        opt = self.optimizers()
        output = self(batch)
        loss = self.loss(batch, output)
        opt.zero_grad()
        self.manual_backward(loss)
        opt.step()
        return loss



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab
#
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

# flake8: noqa

"""
This module provides similar functionality as ``torch.distributed.launch`` with the following
additional functionalities:
1. Worker failures are handled gracefully by restarting all workers.
2. Worker ``RANK`` and ``WORLD_SIZE`` are assigned automatically.
3. Number of nodes is allowed to change between minimum and maximum sizes (elasticity).
**Usage:**
1. Single-node multi-worker
::
    >>> python -m torch.distributed.run
        --standalone
        --nnodes=1
        --nproc_per_node=$NUM_TRAINERS
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)
2. Fault tolerant (fixed sized number of workers, no elasticity):
::
    >>> python -m torch.distributed.run
        --nnodes=$NUM_NODES
        --nproc_per_node=$NUM_TRAINERS
        --rdzv_id=$JOB_ID
        --rdzv_backend=c10d
        --rdzv_endpoint=$HOST_NODE_ADDR
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)
``HOST_NODE_ADDR``, in form <host>[:<port>] (e.g. node1.example.com:29400), specifies the node and
the port on which the C10d rendezvous backend should be instantiated and hosted. It can be any
node in your training cluster, but ideally you should pick a node that has a high bandwidth.
.. note::
   If no port number is specified ``HOST_NODE_ADDR`` defaults to 29400.
3. Elastic (``min=1``, ``max=4``):
::
    >>> python -m torch.distributed.run
        --nnodes=1:4
        --nproc_per_node=$NUM_TRAINERS
        --rdzv_id=$JOB_ID
        --rdzv_backend=c10d
        --rdzv_endpoint=$HOST_NODE_ADDR
        YOUR_TRAINING_SCRIPT.py (--arg1 ... train script args...)
``HOST_NODE_ADDR``, in form <host>[:<port>] (e.g. node1.example.com:29400), specifies the node and
the port on which the C10d rendezvous backend should be instantiated and hosted. It can be any
node in your training cluster, but ideally you should pick a node that has a high bandwidth.
.. note::
   If no port number is specified ``HOST_NODE_ADDR`` defaults to 29400.
**Note on rendezvous backend**:
For multi-node training you need to specify:
1. ``--rdzv_id``: A unique job id (shared by all nodes participating in the job)
2. ``--rdzv_backend``: An implementation of
   :py:class:`torch.distributed.elastic.rendezvous.RendezvousHandler`
3. ``--rdzv_endpoint``: The endpoint where the rendezvous backend is running; usually in form
   ``host:port``.
Currently ``c10d`` (recommended), ``etcd-v2``, and ``etcd`` (legacy)  rendezvous backends are
supported out of the box. To use ``etcd-v2`` or ``etcd``, setup an etcd server with the ``v2`` api
enabled (e.g. ``--enable-v2``).
.. warning::
   ``etcd-v2`` and ``etcd`` rendezvous use etcd API v2. You MUST enable the v2 API on the etcd
   server. Our tests use etcd v3.4.3.
.. warning::
   For etcd-based rendezvous we recommend using ``etcd-v2`` over ``etcd`` which is functionally
   equivalent, but uses a revised implementation. ``etcd`` is in maintenance mode and will be
   removed in a future version.
**Definitions:**
1. ``Node`` - A physical instance or a container; maps to the unit that the job manager works with.
2. ``Worker`` - A worker in the context of distributed training.
3. ``WorkerGroup`` - The set of workers that execute the same function (e.g. trainers).
4. ``LocalWorkerGroup`` - A subset of the workers in the worker group running on the same node.
5. ``RANK`` - The rank of the worker within a worker group.
6. ``WORLD_SIZE`` - The total number of workers in a worker group.
7. ``LOCAL_RANK`` - The rank of the worker within a local worker group.
8. ``LOCAL_WORLD_SIZE`` - The size of the local worker group.
9. ``rdzv_id`` - A user-defined id that uniquely identifies the worker group for a job. This id is
   used by each node to join as a member of a particular worker group.
9. ``rdzv_backend`` - The backend of the rendezvous (e.g. ``c10d``). This is typically a strongly
   consistent key-value store.
10. ``rdzv_endpoint`` - The rendezvous backend endpoint; usually in form ``<host>:<port>``.
A ``Node`` runs ``LOCAL_WORLD_SIZE`` workers which comprise a ``LocalWorkerGroup``. The union of
all ``LocalWorkerGroups`` in the nodes in the job comprise the ``WorkerGroup``.
**Environment Variables:**
The following environment variables are made available to you in your script:
1. ``LOCAL_RANK`` -  The local rank.
2. ``RANK`` -  The global rank.
3. ``GROUP_RANK`` - The rank of the worker group. A number between 0 and ``max_nnodes``. When
   running a single worker group per node, this is the rank of the node.
4. ``ROLE_RANK`` -  The rank of the worker across all the workers that have the same role. The role
   of the worker is specified in the ``WorkerSpec``.
5. ``LOCAL_WORLD_SIZE`` - The local world size (e.g. number of workers running locally); equals to
   ``--nproc_per_node`` specified on ``torch.distributed.run``.
6. ``WORLD_SIZE`` - The world size (total number of workers in the job).
7. ``ROLE_WORLD_SIZE`` - The total number of workers that was launched with the same role specified
   in ``WorkerSpec``.
8. ``MASTER_ADDR`` - The FQDN of the host that is running worker with rank 0; used to initialize
   the Torch Distributed backend.
9. ``MASTER_PORT`` - The port on the ``MASTER_ADDR`` that can be used to host the C10d TCP store.
10. ``TORCHELASTIC_RESTART_COUNT`` - The number of worker group restarts so far.
11. ``TORCHELASTIC_MAX_RESTARTS`` - The configured maximum number of restarts.
12. ``TORCHELASTIC_RUN_ID`` - Equal to the rendezvous ``run_id`` (e.g. unique job id).
**Deployment:**
1. (Not needed for the C10d backend) Start the rendezvous backend server and get the endpoint (to be
   passed as ``--rdzv_endpoint`` to the launcher script)
2. Single-node multi-worker: Start the launcher on the host to start the agent process which
   creates and monitors a local worker group.
3. Multi-node multi-worker: Start the launcher with the same arguments on all the nodes
   participating in training.
When using a job/cluster manager the entry point command to the multi-node job should be this
launcher.
**Failure Modes:**
1. Worker failure: For a training job with ``n`` workers, if ``k<=n`` workers fail all workers
   are stopped and restarted up to ``max_restarts``.
2. Agent failure: An agent failure results in a local worker group failure. It is up to the job
   manager to fail the entire job (gang semantics) or attempt to replace the node. Both behaviors
   are supported by the agent.
3. Node failure: Same as agent failure.
**Membership Changes:**
1. Node departure (scale-down): The agent is notified of the departure, all existing workers are
   stopped, a new ``WorkerGroup`` is formed, and all workers are started with a new ``RANK`` and
   ``WORLD_SIZE``.
2. Node arrival (scale-up): The new node is admitted to the job, all existing workers are stopped,
   a new ``WorkerGroup`` is formed, and all workers are started with a new ``RANK`` and
   ``WORLD_SIZE``.
**Important Notices:**
1. All the items in the important notices section of ``torch.distributed.launch`` apply to this
   module as well.
2. The environment variables necessary to initialize a Torch process group are provided to you by
   this module, no need for you to pass ``RANK`` manually.  To initialize a process group in your
   training script, simply run:
::
 >>> import torch.distributed as dist
 >>> dist.init_process_group(backend="gloo|nccl")
3. On failures or membership changes ALL surviving workers are killed immediately. Make sure to
   checkpoint your progress. The frequency of checkpoints should depend on your job's tolerance
   for lost work.
4. This module only supports homogeneous ``LOCAL_WORLD_SIZE``. That is, it is assumed that all
   nodes run the same number of local workers (per role).
5. ``RANK`` is NOT stable. Between restarts, the local workers on a node can be assgined a
   different range of ranks than before. NEVER hard code any assumptions about the stable-ness of
   ranks or some correlation between ``RANK`` and ``LOCAL_RANK``.
6. When using elasticity (``min_size!=max_size``) DO NOT hard code assumptions about
   ``WORLD_SIZE`` as the world size can change as nodes are allowed to leave and join.
7. It is recommended for your script to have the following structure:
::
  def main():
    load_checkpoint(checkpoint_path)
    initialize()
    train()
  def train():
    for batch in iter(dataset):
      train_step(batch)
      if should_checkpoint:
        save_checkpoint(checkpoint_path)
"""
import logging
import os
import sys
import uuid
from argparse import REMAINDER, ArgumentParser
from typing import Callable, List, Tuple, Union

import torch
from torch.distributed.argparse_util import check_env, env
from torch.distributed.elastic.multiprocessing import Std
from torch.distributed.elastic.rendezvous.utils import _parse_rendezvous_config
from torch.distributed.elastic.utils import macros
from torch.distributed.elastic.utils.logging import get_logger
from torch.distributed.launcher.api import LaunchConfig, elastic_launch


log = get_logger()


def get_args_parser() -> ArgumentParser:
    """Helper function parsing the command line options."""

    parser = ArgumentParser(description="Torch Distributed Elastic Training Launcher")

    #
    # Worker/node size related arguments.
    #

    parser.add_argument(
        "--nnodes",
        action=env,
        type=str,
        default="1:1",
        help="Number of nodes, or the range of nodes in form <minimum_nodes>:<maximum_nodes>.",
    )
    parser.add_argument(
        "--nproc_per_node",
        action=env,
        type=str,
        default="auto",
        help="Number of workers per node; supported values: [auto, cpu, gpu, int].",
    )

    #
    # Rendezvous related arguments
    #

    parser.add_argument(
        "--rdzv_backend",
        action=env,
        type=str,
        default="static",
        help="Rendezvous backend.",
    )
    parser.add_argument(
        "--rdzv_endpoint",
        action=env,
        type=str,
        default="",
        help="Rendezvous backend endpoint; usually in form <host>:<port>.",
    )
    parser.add_argument(
        "--rdzv_id",
        action=env,
        type=str,
        default="none",
        help="User-defined group id.",
    )
    parser.add_argument(
        "--rdzv_conf",
        action=env,
        type=str,
        default="",
        help="Additional rendezvous configuration (<key1>=<value1>,<key2>=<value2>,...).",
    )
    parser.add_argument(
        "--standalone",
        action=check_env,
        help="Start a local standalone rendezvous backend that is represented by a C10d TCP store "
        "on port 29400. Useful when launching single-node, multi-worker job. If specified "
        "--rdzv_backend, --rdzv_endpoint, --rdzv_id are auto-assigned; any explicitly set values "
        "are ignored.",
    )

    #
    # User-code launch related arguments.
    #

    parser.add_argument(
        "--max_restarts",
        action=env,
        type=int,
        default=3,
        help="Maximum number of worker group restarts before failing.",
    )
    parser.add_argument(
        "--monitor_interval",
        action=env,
        type=float,
        default=5,
        help="Interval, in seconds, to monitor the state of workers.",
    )
    parser.add_argument(
        "--start_method",
        action=env,
        type=str,
        default="spawn",
        choices=["spawn", "fork", "forkserver"],
        help="Multiprocessing start method to use when creating workers.",
    )
    parser.add_argument(
        "--role",
        action=env,
        type=str,
        default="default",
        help="User-defined role for the workers.",
    )
    parser.add_argument(
        "-m",
        "--module",
        action=check_env,
        help="Change each process to interpret the launch script as a Python module, executing "
        "with the same behavior as 'python -m'.",
    )
    parser.add_argument(
        "--no_python",
        action=check_env,
        help="Skip prepending the training script with 'python' - just execute it directly. Useful "
        "when the script is not a Python script.",
    )

    parser.add_argument(
        "--run_path",
        action=check_env,
        help="Run the training script with runpy.run_path in the same interpreter."
        " Script must be provided as an abs path (e.g. /abs/path/script.py)."
        " Takes precedence over --no_python.",
    )
    parser.add_argument(
        "--log_dir",
        action=env,
        type=str,
        default=None,
        help="Base directory to use for log files (e.g. /var/log/torch/elastic). The same "
        "directory is re-used for multiple runs (a unique job-level sub-directory is created with "
        "rdzv_id as the prefix).",
    )
    parser.add_argument(
        "-r",
        "--redirects",
        action=env,
        type=str,
        default="0",
        help="Redirect std streams into a log file in the log directory (e.g. [-r 3] redirects "
        "both stdout+stderr for all workers, [-r 0:1,1:2] redirects stdout for local rank 0 and "
        "stderr for local rank 1).",
    )
    parser.add_argument(
        "-t",
        "--tee",
        action=env,
        type=str,
        default="0",
        help="Tee std streams into a log file and also to console (see --redirects for format).",
    )

    #
    # Backwards compatible parameters with caffe2.distributed.launch.
    #

    parser.add_argument(
        "--node_rank",
        type=int,
        action=env,
        default=0,
        help="Rank of the node for multi-node distributed training.",
    )
    parser.add_argument(
        "--master_addr",
        default="127.0.0.1",
        type=str,
        action=env,
        help="Address of the master node (rank 0). It should be either the IP address or the "
        "hostname of rank 0. For single node multi-proc training the --master_addr can simply be "
        "127.0.0.1; IPv6 should have the pattern `[0:0:0:0:0:0:0:1]`.",
    )
    parser.add_argument(
        "--master_port",
        default=29500,
        type=int,
        action=env,
        help="Port on the master node (rank 0) to be used for communication during distributed "
        "training.",
    )

    #
    # Bagua arguments.
    #

    parser.add_argument(
        "--bagua_service_port",
        default=29501,
        type=int,
        help="Port on the master node (rank 0) to used for bagua hyperparameter control during "
        "distributed training.",
    )
    parser.add_argument(
        "--default_bucket_size",
        default=10 * 1024 ** 2,
        type=int,
        help="The `default_bucket_size` variable controls the default size of the bucket used by "
        "Bagua when communicating between pairs of GPUs. Adjusting the bucket size can improve "
        "communication performance. The recommendation is to use powers of 2.",
    )
    parser.add_argument(
        "--autotune_level",
        default=0,
        type=int,
        help="Bagua automatic hyperparameters search level. The higher the level, the larger the "
        "hyperparameter search space, and the longer time it takes. Currently supported levels are 0 and 1.",
    )
    parser.add_argument("--autotune_max_samples", type=int, default=60)
    parser.add_argument("--autotune_sampling_confidence_time", type=float, default=5.0)
    parser.add_argument("--autotune_warmup_time", type=float, default=30.0)
    parser.add_argument(
        "--is_output_autotune_log",
        action="store_true",
        default=False,
        help="Whether autotune output log or not. default is False",
    )
    parser.add_argument(
        "--enable_bagua_net",
        action="store_true",
        default=False,
        help="Enable Bagua-Net optimization for better "
        "communication performance. See https://github.com/BaguaSys/bagua/tree/master/rust/bagua-net"
        " for more details.",
    )

    #
    # Positional arguments.
    #

    parser.add_argument(
        "training_script",
        type=str,
        help="Full path to the (single GPU) training program/script to be launched in parallel, "
        "followed by all the arguments for the training script.",
    )

    # Rest from the training program.
    parser.add_argument("training_script_args", nargs=REMAINDER)

    return parser


def parse_args(args):
    parser = get_args_parser()

    #
    # Legacy arguments.
    #

    parser.add_argument(
        "--use_env",
        default=True,
        action="store_true",
        help="Use environment variable to pass local rank. If set to True (default), the script "
        "will NOT pass --local_rank as argument, and will instead set LOCAL_RANK.",
    )

    return parser.parse_args(args)


def parse_min_max_nnodes(nnodes: str):
    arr = nnodes.split(":")

    if len(arr) == 1:
        min_nodes = max_nodes = int(arr[0])
    elif len(arr) == 2:
        min_nodes = int(arr[0])
        max_nodes = int(arr[1])
    else:
        raise RuntimeError(f'nnodes={nnodes} is not in "MIN:MAX" format')

    return min_nodes, max_nodes


def determine_local_world_size(nproc_per_node: str) -> int:
    try:
        logging.info(f"Using nproc_per_node={nproc_per_node}.")
        return int(nproc_per_node)
    except ValueError:
        if nproc_per_node == "cpu":
            num_proc = os.cpu_count()
            device_type = "cpu"
        elif nproc_per_node == "gpu":
            if not torch.cuda.is_available():
                raise ValueError("Cuda is not available.")
            device_type = "gpu"
            num_proc = torch.cuda.device_count()
        elif nproc_per_node == "auto":
            if torch.cuda.is_available():
                num_proc = torch.cuda.device_count()
                device_type = "gpu"
            else:
                num_proc = os.cpu_count()
                device_type = "cpu"
        else:
            raise ValueError(f"Unsupported nproc_per_node value: {nproc_per_node}")

        log.info(
            f"Using nproc_per_node={nproc_per_node},"
            f" seting to {num_proc} since the instance "
            f"has {os.cpu_count()} {device_type}"
        )
        return int(num_proc)


def get_rdzv_endpoint(args):
    if args.rdzv_backend == "static" and not args.rdzv_endpoint:
        return f"{args.master_addr}:{args.master_port}"
    return args.rdzv_endpoint


def config_from_args(args) -> Tuple[LaunchConfig, Union[Callable, str], List[str]]:
    # If ``args`` not passed, defaults to ``sys.argv[:1]``
    min_nodes, max_nodes = parse_min_max_nnodes(args.nnodes)
    assert 0 < min_nodes <= max_nodes
    assert args.max_restarts >= 0

    nproc_per_node = determine_local_world_size(args.nproc_per_node)
    if "OMP_NUM_THREADS" not in os.environ and nproc_per_node > 1:
        omp_num_threads = 1
        print(
            f"*****************************************\n"
            f"Setting OMP_NUM_THREADS environment variable for each process to be "
            f"{omp_num_threads} in default, to avoid your system being overloaded, "
            f"please further tune the variable for optimal performance in "
            f"your application as needed. \n"
            f"*****************************************"
        )
        # This env variable will be passed down to the subprocesses
        os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)

    rdzv_configs = _parse_rendezvous_config(args.rdzv_conf)

    if args.rdzv_backend == "static":
        rdzv_configs["rank"] = args.node_rank

    rdzv_endpoint = get_rdzv_endpoint(args)

    config = LaunchConfig(
        min_nodes=min_nodes,
        max_nodes=max_nodes,
        nproc_per_node=nproc_per_node,
        run_id=args.rdzv_id,
        role=args.role,
        rdzv_endpoint=rdzv_endpoint,
        rdzv_backend=args.rdzv_backend,
        rdzv_configs=rdzv_configs,
        max_restarts=args.max_restarts,
        monitor_interval=args.monitor_interval,
        start_method=args.start_method,
        redirects=Std.from_str(args.redirects),
        tee=Std.from_str(args.tee),
        log_dir=args.log_dir,
    )

    with_python = not args.no_python
    cmd: Union[Callable, str]
    cmd_args = []
    if args.run_path:
        cmd = run_script_path
        cmd_args.append(args.training_script)
    else:
        if with_python:
            cmd = sys.executable
            cmd_args.append("-u")
            if args.module:
                cmd_args.append("-m")
            cmd_args.append(args.training_script)
        else:
            if not args.use_env:
                raise ValueError(
                    "When using the '--no_python' flag,"
                    " you must also set the '--use_env' flag."
                )
            if args.module:
                raise ValueError(
                    "Don't use both the '--no_python' flag"
                    " and the '--module' flag at the same time."
                )
            cmd = args.training_script
    if not args.use_env:
        log.warning(
            "--use_env is deprecated and will be removed in future releases.\n"
            " Please read local_rank from `os.environ('LOCAL_RANK')` instead."
        )
        cmd_args.append(f"--local_rank={macros.local_rank}")
    cmd_args.extend(args.training_script_args)

    return config, cmd, cmd_args


def run_script_path(training_script: str, *training_script_args: str):
    """
    Runs the provided `training_script` from within this interpreter.
    Usage: `script_as_function("/abs/path/to/script.py", "--arg1", "val1")`
    """
    import runpy
    import sys

    sys.argv = [training_script] + [*training_script_args]
    runpy.run_path(sys.argv[0], run_name="__main__")


def set_bagua_env(args, current_env):
    import pkg_resources

    current_env["BAGUA_SERVICE_PORT"] = str(args.bagua_service_port)
    current_env["BAGUA_DEFAULT_BUCKET_SIZE"] = str(args.default_bucket_size)
    current_env["BAGUA_AUTOTUNE"] = str(args.autotune_level)
    current_env["BAGUA_AUTOTUNE_MAX_SAMPLES"] = str(args.autotune_max_samples)
    current_env["BAGUA_AUTOTUNE_SAMPLING_CONFIDENCE_TIME_S"] = str(
        args.autotune_sampling_confidence_time
    )
    current_env["BAGUA_AUTOTUNE_WARMUP_TIME_S"] = str(args.autotune_warmup_time)
    current_env["BAGUA_IS_OUTPUT_AUTOTUNE_LOG"] = str(int(args.is_output_autotune_log))

    if args.autotune_level > 0:
        current_env["AUTO_TUNE_SERVER_ADDR"] = "{}:{}".format(
            args.master_addr, args.bagua_service_port
        )

    if args.enable_bagua_net:
        current_env["LD_LIBRARY_PATH"] = "{}:{}".format(
            pkg_resources.resource_filename("bagua_core", ".data/bagua-net"),
            current_env["LD_LIBRARY_PATH"],
        )

    # disable flask server warning
    current_env["FLASK_ENV"] = "development"


def run(args):
    set_bagua_env(args, os.environ)

    if args.standalone:
        args.rdzv_backend = "c10d"
        args.rdzv_endpoint = "localhost:29400"
        args.rdzv_id = str(uuid.uuid4())
        log.info(
            f"\n**************************************\n"
            f"Rendezvous info:\n"
            f"--rdzv_backend={args.rdzv_backend} "
            f"--rdzv_endpoint={args.rdzv_endpoint} "
            f"--rdzv_id={args.rdzv_id}\n"
            f"**************************************\n"
        )

    config, cmd, cmd_args = config_from_args(args)
    elastic_launch(
        config=config,
        entrypoint=cmd,
    )(*cmd_args)


def main(args=None):
    args = parse_args(args)
    run(args)


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO, format="[%(levelname)s] %(asctime)s %(module)s: %(message)s"
    )
    log.info(f"Running torch.distributed.run with args: {sys.argv}")
    main()



import logging
import time
import io
import pickle
import multiprocessing
import bagua_core as B
from bagua.service import AutotuneService
from . import env
from .env import (
    get_master_addr,
    get_world_size,
    get_rank,
    get_local_rank,
    get_node_rank,
    get_default_bucket_size,
    get_bagua_service_port,
    get_autotune_server_wait_time,
    find_free_network_port,
)
from enum import IntEnum
from .utils import flatten, unflatten
import torch
import torch.distributed as dist
from bagua.service.autotune_service import AutotuneClient
from functools import lru_cache
from datetime import timedelta
from typing import Optional, List
import torch.distributed.distributed_c10d as c10d
from torch.distributed import ProcessGroup as TorchProcessGroup
import gorilla
import weakref

# fmt: off
__all__ = [
    "ReduceOp", "new_group", "from_torch_group", "init_process_group",
    "is_initialized", "send", "recv", "broadcast", "broadcast_object",
    "reduce", "reduce_inplace", "allreduce", "allreduce_inplace",
    "allgather", "allgather_inplace", "gather", "gather_inplace",
    "scatter", "scatter_inplace", "reduce_scatter", "reduce_scatter_inplace",
    "alltoall", "alltoall_inplace", "barrier", "BaguaProcessGroup"
]

# Process group's global rank to local rank mapping
_pg_group_ranks = {}

# Process group's name to BaguaProcessGroup
_pg_map = {}

# Default process group state
_default_pg = None

# Default store
_default_store = None

# Process group count for default naming
_group_count = 0

# Torch process group to bagua process group
_torch_to_bagua_pg_map = weakref.WeakKeyDictionary({})


# must be consistent with Aluminum ReductionOperator: https://github.com/BaguaSys/Aluminum/blob/master/include/aluminum/base.hpp
class ReduceOp(IntEnum):
    """An enum-like class for available reduction operations: ``SUM``, ``PRODUCT``, ``MIN``, ``MAX``, ``BAND``,
    ``BOR``, ``BXOR`` and ``AVG``."""

    SUM = 0
    PRODUCT = 1
    MIN = 2
    MAX = 3
    BOR = 7
    BAND = 8
    BXOR = 9
    AVG = 10


@gorilla.patches(TorchProcessGroup, filter=lambda name, obj: "bagua" in name)
class BaguaProcessGroupPatch:
    def bagua_patch(self, stream: Optional[torch.cuda.Stream] = None):
        global _torch_to_bagua_pg_map
        if self not in _torch_to_bagua_pg_map:
            _torch_to_bagua_pg_map[self] = from_torch_group(self, stream)

        return self

    @property
    def bagua_pg(self):
        assert self in _torch_to_bagua_pg_map, "cannot find associated Bagua process group in cache, BaguaProcessGroupPatch.bagua_patch(...) needs to be run first to initialize Bagua process group in cache."
        return _torch_to_bagua_pg_map[self]

    def bagua_get_global_communicator(self):
        return get_communicator(self.bagua_pg.group_name, "global")

    def bagua_get_inter_node_communicator(self):
        return get_communicator(self.bagua_pg.group_name, "inter")

    def bagua_get_intra_node_communicator(self):
        return get_communicator(self.bagua_pg.group_name, "intra")


_base = gorilla._get_base(BaguaProcessGroupPatch)
_decorator_data = gorilla.get_decorator_data(_base)
for patch in _decorator_data.patches:
    gorilla.apply(patch)


class BaguaProcessGroup:
    """Definition of Bagua process group."""
    def __init__(self, ranks, stream, group_name):
        self.ranks = ranks
        self.stream = stream
        self.group_name = group_name
        logging.debug(f"Initialize Bagua process group of ranks {self.ranks}")

    def _get_intra_ranks(self):
        rank_mappings = _get_rank_mappings()

        intra_ranks = list(
            filter(
                lambda rank: rank_mappings[rank][0] == get_node_rank(),
                self.ranks,
            )
        )
        return intra_ranks

    def _get_inter_ranks(self):
        rank_mappings = _get_rank_mappings()

        inter_ranks = list(
            filter(
                lambda rank: rank_mappings[rank][1] == rank_mappings[self.ranks[0]][1],
                self.ranks,
            )
        )
        return inter_ranks

    def get_global_communicator(self) -> B.BaguaSingleCommunicatorPy:
        """Returns the global communicator of current process group."""
        return get_communicator(self.group_name, "global")

    def get_inter_node_communicator(self) -> B.BaguaSingleCommunicatorPy:
        """Returns the inter-node communicator of current process group."""
        return get_communicator(self.group_name, "inter")

    def get_intra_node_communicator(self) -> B.BaguaSingleCommunicatorPy:
        """Returns the intra-node communicator of current process group."""
        return get_communicator(self.group_name, "intra")


@lru_cache(maxsize=None)
def _get_rank_mappings():
    rank_mappings = {}

    rank_tensors = torch.cuda.LongTensor(get_world_size(), 2)
    rank_tensors[get_rank()][0] = get_node_rank()
    rank_tensors[get_rank()][1] = get_local_rank()
    allgather_inplace(rank_tensors)

    for i in range(get_world_size()):
        rank_mappings[i] = rank_tensors[i][0].item(), rank_tensors[i][1].item()

    return rank_mappings


def _check_default_pg():
    """
    Helper that checks if the default process group has been initialized, with
    assertion.

    """
    assert _default_pg is not None, "Default process group is not initialized"


def is_initialized():
    """
    Checking if the default process group has been initialized.

    """
    return _default_pg is not None


def _get_default_group():
    """
    Getting the default process group created by :func:`init_process_group`.

    """
    if not is_initialized():
        raise RuntimeError(
            "Default process group has not been initialized, "
            "please make sure to call init_process_group."
        )
    return _default_pg


def _bagua_backend_comm(comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    """
    Return ``None`` if the current process's rank is not in a given communicator.
    Otherwise return the communicator passed in.
    """
    if _rank_not_in_comm(comm):
        return None
    return comm


def new_group(
    ranks: Optional[List[int]] = None, stream: Optional[torch.cuda.Stream] = None
) -> BaguaProcessGroup:
    """
    Creates a new process group.

    This function requires that all processes in the default group (i.e. all
    processes that are part of the distributed job) enter this function, even
    if they are not going to be members of the group. Additionally, groups
    should be created in the same order in all processes.

    Each process group will create three communicators on request, a global communicator,
    a inter-node communicator and a intra-node communicator. Users can access them through
    ``group.get_global_communicator()``, ``group.get_inter_node_communicator()``
    and ``group.get_intra_node_communicator()`` respectively.

    Args:
        ranks: List of ranks of group members. If ``None``, will be
            set to all ranks. Default is ``None``.
        stream: A CUDA stream used to execute NCCL operations. If ``None``,
            CUDA stream of the default group will be used. See
            `CUDA semantics <https://pytorch.org/docs/stable/notes/cuda.html?highlight=stream>`_
            for details.

    Returns:
        A handle of process group that can be given to collective calls.

    .. note::
        The global communicator is used for global communications involving all ranks in the process group.
        The inter-node communicator and the intra-node communicator is used for hierarchical communications
        in this process group.

    .. note::
        For a specific communicator ``comm``, ``comm.rank()`` returns the rank of current process and
        ``comm.nranks()`` returns the size of the communicator.
    """
    global _group_count
    global _pg_group_ranks
    global _pg_map

    _group_count += 1

    if ranks is None:
        ranks = list(range(get_world_size()))
    else:
        # sanity check for the input ranks
        for rank in ranks:
            if rank < 0 or rank >= get_world_size():
                raise ValueError(
                    "Invalid rank {}, should be non-negative and less than world size {}.",
                    rank,
                    get_world_size(),
                )
        ranks = sorted(ranks)

    if stream is None:
        _check_default_pg()
        stream = _get_default_group().stream

    group_name = str(_group_count)
    pg = BaguaProcessGroup(ranks, stream, str(_group_count))
    # Create the global rank to group rank mapping
    _pg_group_ranks[pg] = {
        global_rank: group_rank for group_rank, global_rank in enumerate(ranks)
    }
    _pg_map[group_name] = pg

    return pg


__torch_group_id = 0


def from_torch_group(group, stream: Optional[torch.cuda.Stream] = None) -> BaguaProcessGroup:
    """
    Convert a Pytorch process group to its equivalent Bagua process group.

    Args:
        group: A handle of the Pytorch process group.
        stream: A CUDA stream used to execute NCCL operations. If ``None``,
            CUDA stream of the default group will be used. See :func:`new_group`
            for more information.

    Returns:
       A handle of the Bagua process group.
    """
    global __torch_group_id

    torch_group_id = __torch_group_id
    __torch_group_id += 1

    ranks = None
    if group in c10d._pg_group_ranks:
        ranks = list(c10d._pg_group_ranks[group].keys())
    elif _default_store:
        def rank_key(rank):
            return "global rank of {}.{}".format(torch_group_id, rank)

        _default_store.set(rank_key(group.rank()), env.get_rank())
        ranks = [int(_default_store.get(rank_key(i))) for i in range(group.size())]
    else:
        ranks = list(range(group.size()))

    return new_group(ranks, stream)


@lru_cache(maxsize=None)
def get_communicator(group_name: str, comm_name: str):
    global _pg_map

    pg = _pg_map[group_name]
    if comm_name == "global":
        ranks = pg.ranks
    elif comm_name == "inter":
        ranks = pg._get_inter_ranks()
    elif comm_name == "intra":
        ranks = pg._get_intra_ranks()
    else:
        raise ValueError("comm_name should be one of ['global', 'inter', 'intra']")

    comm_key = "{}_{}_{}".format(group_name, comm_name, ",".join(map(str, ranks)))

    nccl_unique_id = broadcast_nccl_unique_id(comm_key, root=ranks[0])

    if get_rank() not in ranks:
        return CommMember.NON_COMM_MEMBER

    rank = ranks.index(get_rank())
    nranks = len(ranks)

    comm = B.BaguaSingleCommunicatorPy(
        rank=rank,
        nranks=nranks,
        device_id=get_local_rank(),
        stream_ptr=pg.stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )

    logging.debug(
        "init bagua communicator %s-%s ok, global rank: %s rank: %s",
        group_name,
        comm_name,
        get_rank(),
        comm.rank(),
    )
    comm.cuda_stream = pg.stream
    return comm


def _rank_not_in_comm(comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    """
    Return ``True`` if the current process's rank is not in a given communicator.

    """
    if comm is None:
        return False
    return comm == CommMember.NON_COMM_MEMBER


def _rank_not_in_group(group: Optional[BaguaProcessGroup] = None):
    """
    Return ``True`` if the current process is not in a given process group.

    """
    if group is None:
        return False
    return _rank_not_in_comm(group.get_global_communicator())




@lru_cache(maxsize=None)
def get_backend(model_name: str):
    backend = B.BaguaCommBackendPy(100, device_id=get_local_rank())
    backend.model_name = model_name
    return backend


def run_flask_app(port):
    from flask import Flask
    from gevent.pywsgi import WSGIServer
    import os

    os.environ["WERKZEUG_RUN_MAIN"] = "true"

    autotune_service = AutotuneService(
        world_size=get_world_size(),
        autotune_level=env.get_autotune_level(),
        max_samples=env.get_autotune_max_samples(),
        sampling_confidence_time_s=env.get_autotune_sampling_confidence_time_s(),
        warmup_time_s=env.get_autotune_warmup_time_s(),
        is_output_autotune_log=env.get_is_output_autotune_log(),
        default_bucket_size=get_default_bucket_size(),
    )
    app = Flask(__name__)
    app = autotune_service.setup_app(app)

    http_server = WSGIServer(
        listener=("0.0.0.0", port),
        application=app,
        log=None,
    )
    http_server.serve_forever()


_autotune_server = None
_autotune_service_port = None


def start_autotune_server(service_port: int):
    """Starts autotune server in background."""
    global _autotune_server

    _autotune_server = multiprocessing.Process(target=run_flask_app, args=(service_port, ))
    _autotune_server.daemon = True
    _autotune_server.start()


@lru_cache(maxsize=None)
def get_hyperparameters_service_client():
    global _autotune_service_port
    hyperparameters_service_client = AutotuneClient(
        get_master_addr(), _autotune_service_port
    )
    return hyperparameters_service_client


def _find_free_bagua_service_port(store) -> int:
    service_port = get_bagua_service_port()
    if service_port > 0:
        return service_port

    if get_rank() == 0:
        service_port = find_free_network_port()
        store.set("bagua_service_port", str(service_port))
    else:
        service_port = int(store.get("bagua_service_port"))

    return service_port


def init_process_group(store: Optional[torch.distributed.Store] = None):
    """Initializes the PyTorch builtin distributed process group, and this will
    also initialize the distributed package, should be executed before all the
    APIs of Bagua.

    Args:
        store: Key/value store accessible to all workers, used to exchange
            connection/address information. If ``None``, a TCP-based store will be created.
            Default: ``None``.

    Examples::
        >>> import torch
        >>> import bagua.torch_api as bagua
        >>>
        >>> torch.cuda.set_device(bagua.get_local_rank()) # THIS LINE IS IMPORTANT. See the notes below.
        >>> bagua.init_process_group()
        >>>
        >>> model = torch.nn.Sequential(
        ...    torch.nn.Linear(D_in, H),
        ...    torch.nn.ReLU(),
        ...    torch.nn.Linear(H, D_out),
        ...    )
        >>> optimizer = torch.optim.SGD(
        ...    model.parameters(),
        ...    lr=0.01,
        ...    momentum=0.9
        ...    )
        >>> model = model.with_bagua([optimizer], ...)

    .. note::
        Each process should be associated to a CUDA device using `torch.cuda.set_device()`,
        before calling :meth:`init_process_group`. Otherwise you may encounter the
        `fatal runtime error: Rust cannot catch foreign exceptions` error.
    """


    global _default_pg
    global _default_store
    global _autotune_service_port

    if _default_pg is not None:
        raise RuntimeError("trying to initialize the default process group twice!")

    if _default_store is not None:
        raise RuntimeError("The default store has been initialized else where!")

    if store is None:
        timeout = timedelta(minutes=30)
        store, _, _ = next(torch.distributed.rendezvous(url="env://", timeout=timeout))
        store.set_timeout(timeout)
        _default_store = store
    else:
        _default_store = store

    _autotune_service_port = _find_free_bagua_service_port(_default_store)
    if get_rank() == 0 and _autotune_server is None:
        start_autotune_server(_autotune_service_port)

    AUTOTUNE_SERVER_WAIT_TIME = 30
    wait_time = get_autotune_server_wait_time()
    # at least wait 30 seconds
    if wait_time < AUTOTUNE_SERVER_WAIT_TIME:
        wait_time = AUTOTUNE_SERVER_WAIT_TIME

    start = time.time()
    service_ready = False
    while (time.time() - start) < wait_time:
        client = get_hyperparameters_service_client()
        service_ready = client.health_check()
        if service_ready:
            break
    if not service_ready:
        raise Exception("Warning! autotune service not ready after {} seconds. "
                        "You can adjust this duration through "
                        "`BAGUA_AUTOTUNE_SERVER_WAIT_TIME` environment variable.".format(wait_time))

    # TODO remove the dependency on torch process group
    if not dist.is_initialized():
        torch.distributed.init_process_group(
            backend="nccl",
            store=_default_store,
            rank=get_rank(),
            world_size=get_world_size(),
        )  # fmt: off

    _default_pg = new_group(stream=torch.cuda.Stream(priority=-1))


def broadcast_nccl_unique_id(comm_key: str, root):
    global _default_store
    if get_rank() == root:
        idstr = B.BaguaSingleCommunicatorPy.generate_nccl_unique_id_str()
        _default_store.set(comm_key, idstr)
    else:
        idstr = _default_store.get(comm_key)
        idstr = str(idstr, encoding="utf-8")

    return idstr


class comm(object):
    WORLD = object()


class CommMember(object):
    # Alias to group.WORLD for backward compatibility
    WORLD = comm.WORLD
    NON_COMM_MEMBER = object()


def send(tensor: torch.Tensor, dst: int, comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    r"""Sends a tensor to :attr:`dst` synchronously.

    Args:
        tensor: Tensor to send.
        dst: Destination rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.send(tensor.to_bagua_tensor().bagua_backend_tensor(), dst)

    comm.cuda_stream.synchronize()


def recv(tensor: torch.Tensor, src: int, comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    r"""Receives a tensor synchronously.

    Args:
        tensor: Tensor to fill with received data.
        src: Source rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.recv(tensor.to_bagua_tensor().bagua_backend_tensor(), src)

    comm.cuda_stream.synchronize()


def broadcast_coalesced(tensors, src=0, comm: Optional[B.BaguaSingleCommunicatorPy] = None):

    if _rank_not_in_comm(comm):
        return

    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.broadcast(coalesced.to_bagua_tensor().bagua_backend_tensor(), src)
        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    comm.cuda_stream.synchronize()


# Copyright 2020 Uber Technologies, Inc. All Rights Reserved.
# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# This function is copied fron Hovorod: https://github.com/horovod/horovod
# with minor changes.
def broadcast_object(obj: object, src: int = 0, comm: Optional[B.BaguaSingleCommunicatorPy] = None) -> object:
    """Serializes and broadcasts an object from root rank to all other processes.
    Typical usage is to broadcast the ``optimizer.state_dict()``, for example:

        >>> state_dict = broadcast_object(optimizer.state_dict(), 0)
        >>> if get_rank() > 0:
        >>>     optimizer.load_state_dict(state_dict)


    Args:
        obj: An object capable of being serialized without losing any context.
        src: The rank of the process from which parameters will be broadcasted to all other processes.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    Returns:
        The object that was broadcasted from the :attr:`src`.

    .. note::
        This operation will move data to GPU before communication and back to CPU after communication, and it requires
        CPU-GPU synchronization.
    """

    if get_rank() == src:
        b = io.BytesIO()
        pickle.dump(obj, b)
        t = torch.cuda.ByteTensor(bytearray(b.getvalue()))
        # TODO: use IntTensor after int32 communication is supported
        sz = torch.cuda.LongTensor([t.shape[0]])
        broadcast(sz, src, comm)
    else:
        sz = torch.cuda.LongTensor([0])
        broadcast(sz, src, comm)
        t = torch.cuda.ByteTensor(sz.tolist()[0])

    broadcast(t, src, comm)

    if get_rank() != src:
        buf = io.BytesIO(t.cpu().numpy().tobytes())
        obj = pickle.load(buf)

    return obj


def broadcast(tensor: torch.Tensor, src: int = 0, comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    r"""Broadcasts the tensor to all processes associated with the communicator.

    :attr:`tensor` must have the same number of elements in all processes
    participating in the collective.

    Args:
        tensor: Data to be sent if :attr:`src` is the rank of
            current process, and tensor to be used to save received data
            otherwise.
        src: Source rank. Default: 0.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.broadcast(tensor.to_bagua_tensor().bagua_backend_tensor(), src)

    # TODO: remove
    comm.cuda_stream.synchronize()


def reduce(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    dst: int,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    r"""Reduces the tensor data across all processes.

    Only the process whit rank :attr:`dst` is going to receive the final result.

    Args:
        send_tensor: Input of the collective.
        recv_tensor: Output of the collective, must have the same size with :attr:`send_tensor`.
        dst: Destination rank.
        op: One of the values from :class:`ReduceOp`
            enum. Specifies an operation used for element-wise reductions.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            dst,
            int(op),
        )

    comm.cuda_stream.synchronize()


def reduce_inplace(
    tensor: torch.Tensor, dst: int, op: ReduceOp = ReduceOp.SUM, comm: Optional[B.BaguaSingleCommunicatorPy] = None
):
    r"""The in-place version of :func:`reduce`."""

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), dst, int(op)
        )

    comm.cuda_stream.synchronize()


def allreduce_coalesced_inplace(
    tensors,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    if _rank_not_in_comm(comm):
        return

    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.allreduce_inplace(
            coalesced.to_bagua_tensor("allreduce_coalesced"), int(op)
        )

        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    comm.cuda_stream.synchronize()


def allreduce(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """Reduces the tensor data across all processes associated with the communicator in such a way that all get
    the final result. After the call :attr:`recv_tensor` is going to be bitwise identical
    in all processes.

    Args:
        send_tensor: Input of the collective.
        recv_tensor: Output of the collective, must have the same size with :attr:`send_tensor`.
        op: One of the values from :class:`ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.

    Examples::

        >>> from bagua.torch_api import allreduce
        >>>
        >>> # All tensors below are of torch.int64 type.
        >>> # We have 2 process groups, 2 ranks.
        >>> send_tensor = torch.arange(2, dtype=torch.int64, device=tensor.device) + 1 + 2 * rank
        >>> recv_tensor = torch.zeros(2, dtype=torch.int64, device=tensor.device)
        >>> send_tensor
        tensor([1, 2], device='cuda:0') # Rank 0
        tensor([3, 4], device='cuda:1') # Rank 1
        >>> allreduce(send_tensor, recv_tensor)
        >>> recv_tensor
        tensor([4, 6], device='cuda:0') # Rank 0
        tensor([4, 6], device='cuda:1') # Rank 1

        >>> # All tensors below are of torch.cfloat type.
        >>> # We have 2 process groups, 2 ranks.
        >>> send_tensor = torch.tensor([1+1j, 2+2j], dtype=torch.cfloat, device=tensor.device) + 2 * rank * (1+1j)
        >>> recv_tensor = torch.zeros(2, dtype=torch.cfloat, device=tensor.device)
        >>> send_tensor
        tensor([1.+1.j, 2.+2.j], device='cuda:0') # Rank 0
        tensor([3.+3.j, 4.+4.j], device='cuda:1') # Rank 1
        >>> allreduce(send_tensor, recv_tensor)
        >>> recv_tensor
        tensor([4.+4.j, 6.+6.j], device='cuda:0') # Rank 0
        tensor([4.+4.j, 6.+6.j], device='cuda:1') # Rank 1
    """

    if _rank_not_in_comm(comm):
        return

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"


    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allreduce(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            int(op),
        )

    # TODO: remove
    comm.cuda_stream.synchronize()


def allreduce_inplace(
    tensor: torch.Tensor,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`allreduce`."""

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allreduce_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), int(op))

    comm.cuda_stream.synchronize()


def allgather(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """Gathers send tensors from all processes associated with the communicator into :attr:`recv_tensor`.

    Args:
        send_tensor (torch.Tensor): Input of the collective.
        recv_tensor (torch.Tensor): Output of the collective, must have a size of ``comm.nranks * send_tensor.size()`` elements.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allgather(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
        )

    comm.cuda_stream.synchronize()


def allgather_inplace(
    tensor: torch.Tensor,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`allgather`."""

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allgather_inplace(tensor.to_bagua_tensor().bagua_backend_tensor())

    comm.cuda_stream.synchronize()


def gather(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    dst: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """Gathers send tensors from all processes associated with the communicator to :attr:`recv_tensor` in a single process.

    Args:
        send_tensor: Input of the collective.
        recv_tensor: Output of the collective, must have a size of ``comm.nranks * send_tensor.size()`` elements.
        dst: Destination rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.gather(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            dst,
        )

    comm.cuda_stream.synchronize()


def gather_inplace(
    tensor: torch.Tensor,
    count: int,
    dst: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`gather`.

    Args:
        tensor: Input and output of the collective, On the :attr:`dst` rank, it
            must have a size of ``comm.nranks * count`` elements. On non-dst ranks, its size must
            be equal to :attr:``count``.
        count: The per-rank data count to gather.
        dst: Destination rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.gather_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), count, dst)

    comm.cuda_stream.synchronize()


def scatter(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    src: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """Scatters send tensor to all processes associated with the communicator.

    Args:
        send_tensor: Input of the collective, must have a size of ``comm.nranks * recv_tensor.size()`` elements.
        recv_tensor: Output of the collective.
        src: Source rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.scatter(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            src,
        )

    comm.cuda_stream.synchronize()


def scatter_inplace(
    tensor: torch.Tensor,
    count: int,
    src: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`scatter`.

    Args:
        tensor: Input and output of the collective, On the :attr:`src` rank,
            it must have a size of ``comm.nranks * count`` elements. On non-src ranks,
            its size must be equal to :attr:`count`.
        count: The per-rank data count to scatter.
        src: Source rank.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.scatter_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), count, src
        )

    comm.cuda_stream.synchronize()


def reduce_scatter(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """Reduces, then scatters :attr:`send_tensor` to all processes associated with the communicator.

    Args:
        send_tensor (torch.Tensor): Input of the collective, must have a size of ``comm.nranks * recv_tensor.size()`` elements.
        recv_tensor (torch.Tensor): Output of the collective.
        op (ReduceOp, optional): One of the values from :class:`ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_scatter(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            int(op),
        )

    comm.cuda_stream.synchronize()


def reduce_scatter_inplace(
    tensor: torch.Tensor,
    op: ReduceOp = ReduceOp.SUM,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`reduce_scatter`.

    Args:
        tensor (torch.Tensor): Input and output of the collective, the size must be divisible by ``comm.nranks``.
        op (ReduceOp, optional): One of the values from :class:`ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """

    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "send tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_scatter_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), int(op)
        )

    comm.cuda_stream.synchronize()


def alltoall(
    send_tensor: torch.Tensor,
    recv_tensor: torch.Tensor,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """
    Each process scatters :attr:`send_tensor` to all processes associated with the communicator and return the gathered
    data in :attr:`recv_tensor`.

    Args:
        send_tensor (torch.Tensor): Input of the collective, the size must be divisible by ``comm.nranks``.
        recv_tensor (torch.Tensor): Output of the collective, must have equal size with :attr:`send_tensor`.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
        )

    comm.cuda_stream.synchronize()


# TODO combine **inplace API
def alltoall_inplace(
    tensor: torch.Tensor,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`alltoall`."""
    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall_inplace(tensor.to_bagua_tensor().bagua_backend_tensor())

    comm.cuda_stream.synchronize()


def alltoall_v(
    send_tensor: torch.Tensor,
    send_counts: int,
    send_displs: int,
    recv_tensor: torch.Tensor,
    recv_counts: int,
    recv_displs: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """
    Each process scatters :attr:`send_tensor` to all processes associated with the communicator and return the gathered
    data in :attr:`recv_tensor`, each process may send a different amount of data and provide displacements for the input and output data.

    Args:
        send_tensor (torch.Tensor): Input of the collective, the size must be divisible by ``comm.nranks``.
        send_counts: integer array equal to the group size specifying the number of elements to send to each processor.
        send_displs: integer array (of length group size). Entry j specifies the displacement (relative to sendbuf from which to take the outgoing data destined for process j.
        recv_tensor (torch.Tensor): Output of the collective, must have equal size with :attr:`send_tensor`.
        recv_counts: integer array equal to the group size specifying the maximum number of elements that can be received from each processor.
        recv_displs: integer array (of length group size). Entry i specifies the displacement (relative to recvbuf at which to place the incoming data from process i.
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall_v(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            send_counts,
            send_displs,
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_counts,
            recv_displs,
        )

    comm.cuda_stream.synchronize()


def alltoall_v_inplace(
    tensor: torch.Tensor,
    counts: int,
    displs: int,
    comm: Optional[B.BaguaSingleCommunicatorPy] = None,
):
    """The in-place version of :func:`alltoall_v`."""
    if _rank_not_in_comm(comm):
        return

    assert tensor.device != torch.device("cpu"), "recv tensor must be CUDA and dense"

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall_v_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), counts, displs)

    comm.cuda_stream.synchronize()


def barrier(comm: Optional[B.BaguaSingleCommunicatorPy] = None):
    """
    Synchronizes all processes.
    This collective blocks processes until all processes associated with the
    communicator enters this function.

    Args:
        comm: A handle of the Bagua communicator to work on. By default, the global
             communicator of the default process group will be used.
    """
    if _rank_not_in_comm(comm):
        return

    if comm is None or comm is CommMember.WORLD:
        comm = _get_default_group().get_global_communicator()

    event = torch.cuda.current_stream().record_event()
    event.synchronize()

    with torch.cuda.stream(comm.cuda_stream):
        tensor = torch.ones([1], device=torch.cuda.current_device())
        comm.allreduce_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), op=ReduceOp.SUM)

    event = comm.cuda_stream.record_event()
    event.synchronize()



import unittest
from tests.internal.common_utils import find_free_port
import multiprocessing
import os
import torch
import bagua.torch_api as bagua


class MultiProcessTestCase(unittest.TestCase):
    def run_test_locally(self, fn, nprocs, args, results):
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=fn,
                args=(
                    i,
                    nprocs,
                    args,
                    results,
                    env,
                ),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)


def setup_bagua_env(rank, env):
    # initialize subprocess env
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()



import socket


def find_free_port(port=8000, max_port=9000):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while port <= max_port:
        try:
            sock.bind(("", port))
            sock.close()
            return port
        except OSError:
            port += 1
    raise IOError("no free ports")



import os
from setuptools import setup, find_packages
from distutils.errors import (
    DistutilsPlatformError,
)
from setuptools_rust import Binding, RustExtension
import shutil
import sys
from typing import List
from pathlib import Path

library_records = {}


def check_torch_version():
    try:
        import torch
    except ImportError:
        print("import torch failed, is it installed?")

    version = torch.__version__
    if version is None:
        raise DistutilsPlatformError(
            "Unable to determine PyTorch version from the version string '%s'"
            % torch.__version__
        )
    return version


def install_baguanet(destination):
    os.makedirs(destination, exist_ok=True)
    os.system("cd rust/bagua-net/cc && make")
    shutil.move(
        "rust/bagua-net/cc/libnccl-net.so", os.path.join(destination, "libnccl-net.so")
    )


def install_dependency_library():
    nvcc_version = (
        os.popen(
            "nvcc --version | grep release | sed 's/.*release //' |  sed 's/,.*//'"
        )
        .read()
        .strip()
    )
    print("nvcc_version: ", nvcc_version)
    install_baguanet(os.path.join(cwd, "bagua_core", ".data", "bagua-net"))


if __name__ == "__main__":
    import colorama

    colorama.init(autoreset=True)
    cwd = os.path.dirname(os.path.abspath(__file__))

    def check_args(args: List[str]) -> bool:
        for arg in ["build", "install", "develop", "bdist_wheel", "wheel"]:
            if arg in args:
                return True
        return False

    if (
        int(os.getenv("BAGUA_NO_INSTALL_DEPS", 0)) == 0
        and len(sys.argv) > 1
        and check_args(sys.argv)  # noqa: W503
    ):
        print(
            colorama.Fore.BLACK
            + colorama.Back.CYAN
            + "Bagua is automatically installing some system dependencies like bagua-net, to disable set env variable BAGUA_NO_INSTALL_DEPS=1",
        )
        os.system("python3 bagua_core/bagua_install_deps.py")
        install_dependency_library()

    name_suffix = os.getenv("BAGUA_CUDA_VERSION", "")
    if name_suffix != "":
        name_suffix = "-cuda" + name_suffix

    this_directory = Path(__file__).parent
    long_description = (this_directory / "README.md").read_text()

    setup(
        name="bagua" + name_suffix,
        version="0.9.1",
        url="https://github.com/BaguaSys/bagua",
        python_requires=">=3.7",
        description="Bagua is a deep learning training acceleration framework for PyTorch. It provides a one-stop training acceleration solution, including faster distributed training compared to PyTorch DDP, faster dataloader, kernel fusion, and more.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=find_packages(exclude=("tests")),
        package_data={
            "": [
                ".data/bagua-net/libnccl-net.so",
            ]
        },
        rust_extensions=[
            RustExtension(
                "bagua_core.bagua_core",
                path="rust/bagua-core/bagua-core-py/Cargo.toml",
                binding=Binding.PyO3,
                native=False,
            ),
        ],
        author="Kuaishou AI Platform & DS3 Lab",
        author_email="admin@mail.xrlian.com",
        install_requires=[
            "setuptools_rust",
            "colorama",
            "tqdm",
            "deprecation>=2.1",
            "pytest-benchmark>=3.4",
            "scikit-optimize>=0.8.1",
            "scikit-learn>=0.24,<=1.0.1,!=1.0",
            "numpy",
            "flask>=2.0",
            "prometheus_client>=0.11",
            "parallel-ssh==2.9.1",
            "pydantic>=1.8",
            "requests>=2.25",
            "gorilla==0.4.0",
            "gevent>=21.8",
            "xxhash>=2.0",
        ],
        entry_points={
            "console_scripts": [
                "baguarun = bagua.script.baguarun:main",
            ],
        },
        scripts=["bagua/script/bagua_sys_perf", "bagua_core/bagua_install_deps.py"],
        zip_safe=False,
    )



import pickle
from collections import defaultdict
from typing import Callable


__all__ = ["CacheLoader"]


def serialize(input):
    return pickle.dumps(input)


def deserialize(input):
    return pickle.loads(input)


class CacheLoader:
    def __init__(
        self,
        backend: str = "redis",
        dataset_name: str = "",
        writer_buffer_size: int = 1,
        **kwargs,
    ):
        """
        Cache loader caches values calculated by an expensive function by theirs keys via :meth:`get`,
        so that the values can be retrieved faster next time.

        Internally, values are indexed by ``"{dataset_name}_{key}"`` and saved in a distributed key-value
        store, where ``dataset_name`` is specified on initializing, and ``key`` is the argument in :meth:`get`.

        By default, cache loader uses :class:`~bagua.torch_api.contrib.utils.redis_store.RedisStore` as its backend distributed key-value store implementation. It
        supports using a list of existing redis servers or spawning new redis servers. Parameters for :class:`~bagua.torch_api.contrib.utils.redis_store.RedisStore` can be provided here in
        ``**kwargs``.

        Args:
            backend(str): Backend distributed key-value store implementation. Can be ``"redis"``.
            dataset_name(str): Name of the dataset. Default ``""``.
            writer_buffer_size(int): Number of samples to collect before writing to the backend key-value store.
                Useful for improving the backend throughput.

        Example::
            To use a list of existing redis servers for the "redis" backend:

            >>> from bagua.torch_api.contrib import CacheLoader
            >>>
            >>> hosts = [{"host": "192.168.1.0", "port": "7000"}, {"host": "192.168.1.1", "port": "7000"}]
            >>> loader = CacheLoader(backend="redis", hosts=hosts, cluster_mode=True, dataset_name="test")
            >>>
            >>> loader.get(index, lambda x: items[x])

            To spawn new redis servers on training nodes for the "redis" backend, each node with a maximum memory
            limit of 100000000 bytes:

            >>> loader = CacheLoader(backend="redis", hosts=None, cluster_mode=True, capacity_per_node=100000000)

        .. note::
            Cache loaders with the same :attr:`dataset_name` will reuse and overwrite each other's cache.
            Use a different :attr:`dataset_name` if this is not desired.

        """

        self.backend = backend
        self.dataset_name = dataset_name

        if backend == "redis":
            from .utils.redis_store import RedisStore

            self.store = RedisStore(**kwargs)
        else:
            raise ValueError('Invalid backend, only support "redis" currently')

        self.fetcher = BatchFetcher(self.store, 1, writer_buffer_size)

    def get(self, key: str, load_fn: Callable[[str], None]):
        """
        Returns the value associated with :attr:`key` in cache, use  :attr:`load_fn` to create the entry if the key does not exist
        in the cache. :attr:`load_fn` is a function taking :attr:`key` as its argument, and returning corresponding value to
        be cached.
        """

        cache_key = "{}_{}".format(self.dataset_name, key)
        ret = self.fetcher.read(cache_key)

        if ret is None:
            ret = load_fn(key)
            # write to store
            self.fetcher.write(cache_key, ret)
        return ret

    def num_keys(self):
        """Returns the number of keys in the cache."""

        return self.store.num_keys()


class BatchFetcher:
    def __init__(self, store, read_buffer_size, writer_buffer_size):
        self.store = store
        self.read_buffer_size = max(1, read_buffer_size)
        self.writer_buffer_size = max(1, writer_buffer_size)

        self.write_map = defaultdict()
        self.write_cnt = 0
        self.read_cnt = 0

        self.last_write_tms = None

    def read(self, key):
        self.read_cnt += 1

        try:
            ret = self.store.get(key)
        except Exception:
            ret = None
        else:
            self.write_post_read()

        if ret is not None:
            return deserialize(ret)
        return ret

    def write(self, key, value):
        self.write_cnt += 1

        self.write_map[key] = serialize(value)
        if self.write_cnt % self.writer_buffer_size == 0:
            self.flush_write_map()

    def write_post_read(self):
        if self.read_cnt % 1000 == 0 and len(self.write_map) > 0:
            self.flush_write_map()

    def flush_write_map(self):
        try:
            self.store.mset(self.write_map)
        except Exception:
            pass
        else:
            self.write_map.clear()



# TODO: @shjwudp merge with service module
import copy
import os
import re
import time
from pssh.clients import ParallelSSHClient
from pssh.exceptions import Timeout


from .bayesian_optimizer import (
    IntParam,
    BayesianOptimizer,
)


def sysperf(host_list, nproc_per_node, ssh_port, env: dict = {}):
    assert len(host_list) != 0, "Invalid host_list={}".format(host_list)

    if "PATH" not in env:
        env["PATH"] = os.environ["PATH"]

    pretreat_cmd = [
        "shopt -s huponexit &&",
    ]
    for k, v in env.items():
        pretreat_cmd.append(
            "{key}={value} &&".format(
                key=k,
                value=v,
            )
        )

    master_addr = host_list[0]
    host_args = []
    for i, _ in enumerate(host_list):
        host_args.append(
            {
                "cmd": " ".join(
                    pretreat_cmd
                    + [
                        "python -m bagua.distributed.launch",
                        "--nproc_per_node={}".format(nproc_per_node),
                        "--nnodes={} --node_rank={}".format(len(host_list), i),
                        '--master_addr="{}"'.format(master_addr),
                        "--master_port={}".format(8124),
                        "$(which bagua_sys_perf) --model vgg16",
                    ]
                ),
            }
        )

    client = ParallelSSHClient(host_list, port=ssh_port)
    output = client.run_command(
        "%(cmd)s",
        host_args=host_args,
        shell="bash -xc",
        use_pty=True,  # The key configuration of process safe exit
        read_timeout=60,
    )
    speed_pattern = re.compile(
        r"Total img/sec on (\d+) (\S+)\(s\): (\d*\.\d+|\d+) \+-(\d*\.\d+|\d+)"
    )
    host_out = output[0]
    m = None

    st = time.time()
    try:
        for line in host_out.stdout:
            print(line, flush=True)
            m = speed_pattern.search(line)
            if m:
                break
    except Timeout:
        print("Timeout 1, spend={}".format(time.time() - st))
        pass

    if m is None:
        return (None, None, 0.0, None)

    assert m, "no speed pattern, host_out.exit_code={}, host_out.stderr={}".format(
        host_out.exit_code, list(host_out.stderr)
    )

    ngpus = int(m.groups()[0])
    device = m.groups()[1]
    speed = float(m.groups()[2])
    speed_std = float(m.groups()[3])

    return (ngpus, device, speed, speed_std)


def autotune_system_hyperparameters(host_list, nproc_per_node, ssh_port):
    def _sysperf(env={}):
        result = sysperf(host_list, nproc_per_node, ssh_port, env=env)
        print(result)
        return result

    optim = BayesianOptimizer(
        {
            "NCCL_MIN_NCHANNELS": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    12,
                ),
            ),
            "NCCL_SOCKET_NTHREADS": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    8,
                ),
            ),
            "NCCL_NSOCKS_PERTHREAD": IntParam(
                val=0,  # 0 means no set
                space_dimension=(
                    0,
                    8,
                ),
            ),
            "nccl_buffsize_2p": IntParam(
                val=0,  # power of 2, 0 means no set
                space_dimension=(
                    0,
                    26,
                ),
            ),
        }
    )

    param_dict = {
        "NCCL_MIN_NCHANNELS": 0,
        "NCCL_SOCKET_NTHREADS": 0,
        "NCCL_NSOCKS_PERTHREAD": 0,
        "nccl_buffsize_2p": 0,
    }

    result_list = []
    for i in range(100):
        env_vars = copy.deepcopy(param_dict)
        for k, v in list(env_vars.items()):
            if v == 0:
                del env_vars[k]

        if "nccl_buffsize_2p" in env_vars:
            env_vars["NCCL_BUFFSIZE"] = 2 ** env_vars["nccl_buffsize_2p"]

        (_, _, speed, speed_std) = _sysperf(env=env_vars)
        result_list.append([copy.deepcopy(env_vars), speed, speed_std])
        optim.tell(param_dict, speed)
        param_dict = optim.ask()

    result_list = sorted(result_list, key=lambda x: -x[1])
    print(result_list)

    result_reduct = {}
    for (setting, speed, _) in result_list:
        key = tuple(sorted(setting.items()))
        if key not in result_reduct:
            result_reduct[key] = []
        result_reduct[key].append(speed)
    result_reduct = [
        [setting, sum(speed_list) / len(speed_list)]
        for (setting, speed_list) in result_reduct.items()
    ]
    result_reduct = sorted(result_reduct, key=lambda item: -item[1])
    print(result_reduct)

    return result_reduct[0][0]



# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab
#
# All rights reserved.

import logging
import os
import re
import sys
import torch
import torch.distributed as dist
from bagua.torch_api import BaguaModule
from bagua.torch_api.model_parallel.moe import MoE
from collections import defaultdict
from typing import Dict, Optional, Tuple


def _has_moe_layers(model: BaguaModule) -> Tuple[bool, int]:
    bagua_has_moe_layers = False
    bagua_moe_num_experts = 0
    for name, module in model.named_modules():
        if isinstance(module, MoE):
            bagua_has_moe_layers = True
            bagua_moe_num_experts = module.num_experts
            break
    return bagua_has_moe_layers, bagua_moe_num_experts


def _ensure_directory_exists(filename: str):
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)


def _get_optimizer_ckpt_name(
    checkpoints_path: str,
    iteration: int,
    expert_parallel_rank: int,
    mp_rank: Optional[int] = 0,
    release: Optional[bool] = False,
) -> str:
    if release:
        directory = "release"
    else:
        directory = "iter_{:07d}".format(iteration)
    ckpt_name = os.path.join(
        checkpoints_path,
        directory,
        f"expert_parallel_rank_{expert_parallel_rank}_mp_rank_{mp_rank:02d}_optim_states.pt",
    )
    return ckpt_name


def _get_expert_ckpt_name(
    checkpoints_path: str,
    expert_id: str,
    iteration: int,
    mp_rank: Optional[int] = 0,
    release: Optional[bool] = False,
) -> str:
    if release:
        directory = "release"
    else:
        directory = "iter_{:07d}".format(iteration)
    ckpt_name = os.path.join(
        checkpoints_path,
        directory,
        f"expert_{expert_id}_mp_rank_{mp_rank:02d}_model_states.pt",
    )
    return ckpt_name


def _get_model_ckpt_name(
    checkpoints_path: str,
    iteration: int,
    mp_rank: Optional[int] = 0,
    release: Optional[bool] = False,
) -> str:
    if release:
        directory = "release"
    else:
        directory = "iter_{:07d}".format(iteration)
    return os.path.join(
        checkpoints_path, directory, f"mp_rank_{mp_rank:02d}_model_states.pt"
    )


def _get_checkpoint_tracker_filename(checkpoints_path: str) -> str:
    return os.path.join(checkpoints_path, "latest_checkpointed_iteration.txt")


def _read_metadata(tracker_filename: str) -> Tuple[int, bool]:
    iteration = 0
    release = False
    with open(tracker_filename, "r") as f:
        metastring = f.read().strip()
        try:
            iteration = int(metastring)
        except ValueError:
            release = metastring == "release"
            if not release:
                logging.error(
                    "Invalid metadata file {}. Exiting".format(tracker_filename)
                )
                sys.exit()
    assert iteration >= 0 or release, "error parsing metadata file {}".format(
        tracker_filename
    )

    return iteration, release


def save_checkpoint(
    iteration: int,
    checkpoints_path: str,
    model: BaguaModule,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
):
    """Save model checkpoint.

    Args:
        iteration(int): Training Iteration.
        checkpoints_path(str): Path of checkpoints.
        model(BaguaModule): The model to save.
        optimizer(torch.optim.Optimizer, optional): The optimizer to save. Default: ``None``.
        lr_scheduler(torch.optim.lr_scheduler._LRScheduler, optional): The LR scheduler to save. Default: ``None``.
    """
    logging.info(
        "saving checkpoint at iterration {:7d} to {}".format(
            iteration, checkpoints_path
        )
    )

    bagua_has_moe_layers, bagua_moe_num_experts = _has_moe_layers(model)
    if bagua_has_moe_layers:
        _save_moe_checkpoint(
            iteration,
            checkpoints_path,
            bagua_moe_num_experts,
            model,
            optimizer,
            lr_scheduler,
        )
    else:
        _save_checkpoint(iteration, checkpoints_path, model, optimizer, lr_scheduler)

    logging.info(
        "successfully saved checkpoint at iterration {:7d} to {}".format(
            iteration, checkpoints_path
        )
    )

    # update the latest iteration
    if not dist.is_initialized() or dist.get_rank() == 0:
        tracker_filename = _get_checkpoint_tracker_filename(checkpoints_path)
        with open(tracker_filename, "w") as f:
            f.write(str(iteration))

    if dist.is_initialized():
        dist.barrier()


def _save_checkpoint(
    iteration: int,
    checkpoints_path: str,
    model: BaguaModule,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
):
    if not dist.is_initialized() or dist.get_rank() == 0:
        state_dict = {}
        state_dict["iteration"] = iteration
        state_dict["model"] = model.state_dict()
        if optimizer is not None:
            state_dict["optimizer"] = optimizer.state_dict()
        if lr_scheduler is not None:
            state_dict["lr_scheduler"] = lr_scheduler.state_dict()

        checkpoint_name = _get_model_ckpt_name(checkpoints_path, iteration)
        _ensure_directory_exists(checkpoint_name)
        torch.save(state_dict, checkpoint_name)


def _save_moe_checkpoint(
    iteration: int,
    checkpoints_path: str,
    num_experts: int,
    model: BaguaModule,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
):
    world_size = 1 if not dist.is_initialized() else dist.get_world_size()
    expert_parallel_rank = 0 if not dist.is_initialized() else dist.get_rank()
    num_local_experts = num_experts // world_size
    experts_state_dict, model_state_dict = _get_moe_state_dict(
        model.state_dict(), num_local_experts, expert_parallel_rank
    )

    #  Each rank saves its local experts
    for global_expert_id, expert_state_dict in experts_state_dict.items():
        expert_save_dir = _get_expert_ckpt_name(
            checkpoints_path, global_expert_id, iteration
        )
        logging.info(
            f"Saving model expert {global_expert_id} checkpoint: {expert_save_dir}"
        )
        _ensure_directory_exists(expert_save_dir)
        torch.save(expert_state_dict, expert_save_dir)

    # Save optimizer states. They are different across each exp parallel rank.
    optimizer_state = {"optimizer": optimizer.state_dict() if optimizer else None}
    torch.save(
        optimizer_state,
        _get_optimizer_ckpt_name(checkpoints_path, iteration, expert_parallel_rank),
    )

    if expert_parallel_rank == 0:
        state_dict = {}
        state_dict["iteration"] = iteration
        state_dict["model"] = model_state_dict
        if lr_scheduler is not None:
            state_dict["lr_scheduler"] = lr_scheduler.state_dict()

        # Save.
        checkpoint_name = _get_model_ckpt_name(checkpoints_path, iteration)
        _ensure_directory_exists(checkpoint_name)
        torch.save(state_dict, checkpoint_name)


def _get_moe_state_dict(
    full_state_dict: Dict[str, torch.Tensor],
    num_local_experts: int,
    expert_parallel_rank: int,
):
    experts_state_dict, moe_state_dict = defaultdict(dict), {}
    for key in list(full_state_dict.keys()):
        if "expert" in key and "moe.gate.wg.weight" not in key:
            moe_state_dict[key] = full_state_dict.pop(key)
    non_moe_state_dict = full_state_dict

    moe_str_prefix = ".bagua_moe.experts.bagua_experts."
    for key in list(moe_state_dict.keys()):
        m = re.match(f".*{moe_str_prefix}([0-9]+).*", key)
        local_expert_id = None
        if not m:
            logging.warning(f"No expert found in key {key}.")
        else:
            local_expert_id = m.group(1)

        global_expert_id = expert_parallel_rank * num_local_experts + int(
            local_expert_id
        )
        expert_key = key.replace(
            f"{moe_str_prefix}{local_expert_id}", f"{moe_str_prefix}{global_expert_id}"
        )
        experts_state_dict[str(global_expert_id)][expert_key] = moe_state_dict.pop(key)

    return experts_state_dict, non_moe_state_dict


def load_checkpoint(
    checkpoints_path: str,
    model: BaguaModule,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
    strict: Optional[bool] = True,
):
    """Load a model checkpoint and return the iteration.

    Args:
        checkpoints_path(str): Path of checkpoints.
        model(BaguaModule): The model to load on.
        optimizer(torch.optim.Optimizer, optional): The optimizer to load on. Default: ``None``.
        lr_scheduler(torch.optim.lr_scheduler._LRScheduler, optional): The LR scheduler to load on. Default: ``None``.
        strict (bool, optional): whether to strictly enforce that the keys in
            ``state_dict`` of the checkpoint match the keys returned by this module's
            state_dict() function. Default: ``True``.
    """

    tracker_filename = _get_checkpoint_tracker_filename(checkpoints_path)

    # If no tracker file, return iretation zero.
    if not os.path.isfile(tracker_filename):
        logging.warning(f"could not find checkpoint metadata file {tracker_filename}")
        return 0

    iteration, release = _read_metadata(tracker_filename)
    logging.info(f"loading checkpoint from {checkpoints_path} at iteration {iteration}")

    _load_checkpoint(iteration, checkpoints_path, model, optimizer, lr_scheduler)

    if dist.is_initialized():
        dist.barrier()

    logging.info(
        f"successfully loaded checkpoint from {checkpoints_path} at {iteration}"
    )
    return iteration


def _load_checkpoint(
    iteration: int,
    checkpoints_path: str,
    model: BaguaModule,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
    strict: Optional[bool] = True,
):
    expert_parallel_rank = 0 if not dist.is_initialized() else dist.get_rank()
    checkpoint_name = _get_model_ckpt_name(checkpoints_path, iteration)

    model_checkpoint = torch.load(checkpoint_name, map_location="cpu")
    bagua_has_moe_layers, bagua_moe_num_experts = _has_moe_layers(model)
    if bagua_has_moe_layers:
        num_local_experts = bagua_moe_num_experts // dist.get_world_size()
        _load_moe_state_dict(
            checkpoints_path,
            iteration,
            num_local_experts,
            expert_parallel_rank,
            state_dict=model_checkpoint["model"],
        )

    if bagua_has_moe_layers and optimizer is not None:
        optim_load_path = _get_optimizer_ckpt_name(
            checkpoints_path, iteration, expert_parallel_rank
        )
        optim_checkpoint = torch.load(optim_load_path, map_location=torch.device("cpu"))
    else:
        optim_checkpoint = model_checkpoint

    model.load_state_dict(model_checkpoint["model"], strict=strict)

    # Optimizer.
    if optimizer is not None:
        optimizer.load_state_dict(optim_checkpoint["optimizer"])
    if lr_scheduler is not None:
        lr_scheduler.load_state_dict(model_checkpoint["lr_scheduler"])


def _load_moe_state_dict(
    checkpoints_path: str,
    iteration: int,
    num_local_experts: int,
    expert_parallel_rank: int,
    state_dict: Dict[str, torch.Tensor],
):
    for local_expert_id in range(num_local_experts):
        global_expert_id = expert_parallel_rank * num_local_experts + local_expert_id
        expert_state_dict = torch.load(
            _get_expert_ckpt_name(checkpoints_path, str(global_expert_id), iteration),
            map_location=torch.device("cpu"),
        )

        # Updating global -> local expert ids
        moe_str_prefix = ".bagua_moe.experts.bagua_experts."
        for key in list(expert_state_dict.keys()):
            local_key = key.replace(
                f"{moe_str_prefix}{global_expert_id}",
                f"{moe_str_prefix}{local_expert_id}",
            )
            expert_state_dict[local_key] = expert_state_dict.pop(key)
        state_dict.update(expert_state_dict)



import pytest
import torch
from tests import skip_if_cuda_not_available

if torch.cuda.is_available():
    from pytorch_lightning import Trainer
    from pytorch_lightning.strategies import BaguaStrategy
    from tests.pytorch_lightning.boring_model import BoringModel
else:
    Trainer = None
    BaguaStrategy = None
    BoringModel = object


class TestModel(BoringModel):
    def __init__(self):
        super().__init__()
        self.layer = torch.nn.Linear(32, 32)


class TestModel4QAdam(TestModel):
    def configure_optimizers(self):
        from bagua.torch_api.algorithms.q_adam import QAdamOptimizer

        optimizer = QAdamOptimizer(self.layer.parameters(), lr=0.05, warmup_steps=20)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1)
        return [optimizer], [lr_scheduler]


@skip_if_cuda_not_available()
def test_bagua_default(tmpdir):
    torch.manual_seed(13)
    model = TestModel()
    assert torch.norm(model.layer.weight) == 3.1995556354522705
    trainer = Trainer(
        default_root_dir=tmpdir,
        max_epochs=1,
        strategy="bagua",
        accelerator="gpu",
        devices=1,
    )
    trainer.fit(model)
    trainer.test(model)
    assert torch.norm(model.layer.weight) == 2.4819390773773193


@pytest.mark.parametrize(
    ["algorithm", "criterion"],
    [
        ("gradient_allreduce", 2.835376262664795),
        ("bytegrad", 2.835047960281372),
        ("decentralized", 2.835376262664795),
        ("low_precision_decentralized", 2.8350701332092285),
    ],
)
@skip_if_cuda_not_available()
def test_bagua_algorithm(tmpdir, algorithm, criterion):
    torch.manual_seed(13)
    model = TestModel()
    assert torch.norm(model.layer.weight) == 3.1995556354522705
    bagua_strategy = BaguaStrategy(algorithm=algorithm)
    trainer = Trainer(
        default_root_dir=tmpdir,
        max_epochs=1,
        strategy=bagua_strategy,
        accelerator="gpu",
        devices=2,
    )
    trainer.fit(model)
    trainer.test(model)
    assert torch.norm(model.layer.weight) == criterion


@skip_if_cuda_not_available()
def test_bagua_async(tmpdir):
    model = TestModel()
    bagua_strategy = BaguaStrategy(
        algorithm="async", warmup_steps=10, sync_interval_ms=10
    )
    trainer = Trainer(
        default_root_dir=tmpdir,
        max_epochs=1,
        strategy=bagua_strategy,
        accelerator="gpu",
        devices=2,
    )
    trainer.fit(model)
    trainer.test(model)
    assert torch.norm(model.layer.weight) < 5


@skip_if_cuda_not_available()
def test_qadam(tmpdir):
    torch.manual_seed(13)
    model = TestModel4QAdam()
    assert torch.norm(model.layer.weight) == 3.1995556354522705
    bagua_strategy = BaguaStrategy(algorithm="qadam")
    trainer = Trainer(
        default_root_dir=tmpdir,
        max_epochs=1,
        strategy=bagua_strategy,
        accelerator="gpu",
        devices=2,
    )
    trainer.fit(model)
    trainer.test(model)
    assert torch.norm(model.layer.weight) == 6.891299724578857



#!/usr/bin/env python3
from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.data_parallel.bagua_distributed import BaguaDistributedDataParallel
from bagua.torch_api.algorithms import Algorithm, AlgorithmImpl
from bagua.torch_api.communication import (
    new_group,
    barrier,
    _pg_group_ranks,
    BaguaProcessGroup,
)
from typing import Union, List
from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api.env import get_rank
from enum import IntEnum
import bagua
import threading
import time
import torch
import logging
from concurrent.futures import ThreadPoolExecutor, wait


__all__ = ["AsyncModelAverageAlgorithm", "AsyncModelAverageAlgorithmImpl"]


class _AsyncInternalState(IntEnum):
    NEW = 0
    SCHEDULED = 1
    STARTED = 2
    STOPPED = 3


class AsyncModelAverageAlgorithmImpl(AlgorithmImpl):
    def __init__(
        self,
        process_group: BaguaProcessGroup,
        peer_selection_mode: str = "all",
        sync_interval_ms: int = 500,
        warmup_steps: int = 0,
    ):
        """
        Implementation of the
        `AsyncModelAverage <https://tutorials.baguasys.com/algorithms/async-model-average.html>`_
        algorithm.

        The asynchronous implementation is experimental, and imposes some restrictions.
        With such asynchronous algorithm, the number of iterations on each worker are different. Therefore
        the current implementation assumes that the dataset is an endless stream, and all workers continuously
        synchronize between each other.

        Users should call :meth:`abort` to manually stop the algorithm's continuous synchronization process.
        For example, for a model wrapped with `.with_bagua(...)`, you can abort with `model.bagua_algorithm.abort(model)`,
        and resume with `model.bagua_algorithm.resume(model)`.

        Args:
            process_group (BaguaProcessGroup): The process group to work on.
            peer_selection_mode (str): The way how workers communicate with each other. Currently ``"all"`` is supported.
                ``"all"`` means all workers' weights are synchronized during each communication.
            sync_interval_ms (int): Number of milliseconds between model synchronizations.
            warmup_steps (int): Number of steps to warm up by doing gradient allreduce before doing asynchronous
                model averaging. Use 0 to disable.
        """

        super(AsyncModelAverageAlgorithmImpl, self).__init__(process_group)
        self.peer_selection_mode = peer_selection_mode
        self.sync_interval_ms = sync_interval_ms
        self.step_id = 0
        self.warmup_steps = warmup_steps

        self.cuda_event = torch.cuda.Event()

        self.executor = ThreadPoolExecutor(max_workers=1)
        self.cv = threading.Condition()
        self.notified = False

        self.status = _AsyncInternalState.NEW

        # process group used for background communication
        process_ranks = list(_pg_group_ranks[self.process_group])
        self.thread_group = new_group(
            process_ranks, stream=torch.cuda.Stream(priority=-1)
        )

    def tensors_to_buckets(
        self, tensors: List[List[BaguaTensor]], do_flatten: bool
    ) -> List[BaguaBucket]:
        # TODO: async algorithm conflict with fused optimizer, can only support flattened inplace bucket.
        assert do_flatten, "Async algorithm supports `do_flatten=True` only"
        if self.step_id < self.warmup_steps:
            return super().tensors_to_buckets(tensors, do_flatten)

        all_tensors = []
        for idx, bucket in enumerate(tensors):
            all_tensors.extend(bucket)

        bagua_bucket = BaguaBucket(all_tensors, flatten=do_flatten, name=str(0))

        return [bagua_bucket]

    def init_tensors(
        self, bagua_ddp: BaguaDistributedDataParallel
    ) -> List[BaguaTensor]:
        parameters = bagua_ddp.bagua_build_params()
        tensors = []
        for name, param in parameters.__reversed__():
            if self.step_id < self.warmup_steps:
                param = param.bagua_ensure_grad().ensure_bagua_tensor(
                    name,
                    bagua_ddp.bagua_module_name,
                    getter_closure=lambda param: param.grad,
                    setter_closure=lambda param, t: setattr(param, "grad", t),
                )
                tensors.append(param)
            else:
                p = param.ensure_bagua_tensor(name, bagua_ddp.bagua_module_name)
                tensors.append(p)

        return tensors

    def init_forward_pre_hook(self, bagua_ddp: BaguaDistributedDataParallel):
        def hook(input):
            if (
                self.step_id > self.warmup_steps
                and self.sync_interval_ms > 0  # noqa: W503
            ):
                if self.status == _AsyncInternalState.NEW:
                    self.future = self.executor.submit(self._run_async_loop, bagua_ddp)
                    self.status = _AsyncInternalState.SCHEDULED
                elif (
                    self.status == _AsyncInternalState.SCHEDULED
                    and self.future.running()
                ):
                    with self.cv:
                        self.notified = True
                        self.cv.notify()

                    self.status = _AsyncInternalState.STARTED

                self._lock_model(bagua_ddp)

        return hook

    def init_backward_hook(self, bagua_ddp: BaguaDistributedDataParallel):
        def hook(parameter_name, parameter):
            if self.step_id <= self.warmup_steps:
                parameter.bagua_mark_communication_ready()

        return hook

    def init_post_backward_hook(self, bagua_ddp: BaguaDistributedDataParallel):
        def hook():
            if self.step_id <= self.warmup_steps:
                bagua_ddp._bagua_backend.wait_pending_comm_ops()
            else:
                self._unlock_model(bagua_ddp)

        return hook

    def need_reset(self):
        self.step_id += 1

        if self.warmup_steps > 0 and self.step_id == self.warmup_steps + 1:
            logging.info(f"Async model average starts from step {self.step_id}")
            return True
        else:
            return False

    def init_operations(
        self,
        bagua_ddp: BaguaDistributedDataParallel,
        bucket: BaguaBucket,
    ):
        bagua_ddp._bagua_backend.wait_pending_comm_ops()
        bucket.clear_ops()

        if self.step_id < self.warmup_steps:
            bucket.append_centralized_synchronous_op(
                hierarchical=False,
                average=True,
                group=self.thread_group,
            )
        else:
            async_op = bucket.append_asynchronous_model_average_op(
                peer_selection_mode=self.peer_selection_mode, group=self.thread_group
            )
            bucket._async_op = async_op

    def _lock_model(self, bagua_ddp: BaguaDistributedDataParallel):
        torch.cuda.current_stream().record_event(self.cuda_event)
        self.cuda_event.synchronize()

        for bucket in bagua_ddp.bagua_buckets:
            bucket._async_op.lock_weight()

    def _unlock_model(self, bagua_ddp: BaguaDistributedDataParallel):
        torch.cuda.current_stream().record_event(self.cuda_event)
        self.cuda_event.synchronize()

        for bucket in bagua_ddp.bagua_buckets:
            bucket._async_op.unlock_weight()

    def _check_op_status(self, bagua_ddp: BaguaDistributedDataParallel):
        return (
            hasattr(bagua_ddp.bagua_buckets[0], "_async_op")
            and bagua_ddp.bagua_buckets[0]._async_op.get_status()
        )

    def _run_async_loop(self, bagua_ddp: BaguaDistributedDataParallel):
        with self.cv:
            while not self.notified:
                self.cv.wait()

        comm_step = 0
        while self._check_op_status(bagua_ddp):
            start_time = time.time()

            for bucket in bagua_ddp.bagua_buckets:
                for tensor in bucket.tensors:
                    tensor.bagua_mark_communication_ready_without_synchronization()

            bagua_ddp._bagua_backend.wait_pending_comm_ops()
            duration = (time.time() - start_time) * 1000

            logging.debug(
                "Process {} async communication cost {}ms, comm_step={}".format(
                    get_rank(), duration, comm_step
                )
            )
            comm_step += 1
            time.sleep(self.sync_interval_ms / 1000)

    def abort(
        self,
        bagua_ddp: "Union[BaguaDistributedDataParallel, bagua.torch_api.distributed.BaguaModule]",
    ):
        """
        Stop background asynchronous communications. Should be called after
        training.

        Args:
            bagua_ddp (Union[BaguaDistributedDataParallel, bagua.torch_api.distributed.BaguaModule]): Bagua distributed data parallel module.
        """

        # TODO: Remove the following code when BaguaModule is completely obsolete
        if type(bagua_ddp) is BaguaDistributedDataParallel:
            pass
        elif hasattr(bagua_ddp, "inner"):
            # Is bagua.torch_api.data_parallel.DistributedDataParallel
            bagua_ddp = bagua_ddp.inner
        elif hasattr(bagua_ddp, "bagua_ddp"):
            # Is bagua.torch_api.distributed.BaguaModule
            bagua_ddp = bagua_ddp.bagua_ddp
        else:
            raise Exception(
                "Unexpect input bagua_ddp({}), it should be BaguaDistributedDataParallel or bagua.torch_api.distributed.BaguaModule.".format(
                    type(bagua_ddp)
                )
            )

        if self.status in [_AsyncInternalState.SCHEDULED, _AsyncInternalState.STARTED]:
            barrier(comm=self.process_group.get_global_communicator())
            if hasattr(bagua_ddp.bagua_buckets[0], "_async_op"):
                bagua_ddp.bagua_buckets[0]._async_op.abort()
            with self.cv:
                self.notified = True
                self.cv.notify()

            wait([self.future])  # pytype: disable=attribute-error
            self.status = _AsyncInternalState.STOPPED
            logging.debug("Process {} async communication aborted.".format(get_rank()))

    def resume(
        self,
        bagua_ddp: "Union[BaguaDistributedDataParallel, bagua.torch_api.distributed.BaguaModule]",
    ):
        """
        Resume aborted background asynchronous communications (see :meth:`abort`). Should be called before training.

        Args:
            bagua_ddp (Union[BaguaDistributedDataParallel, bagua.torch_api.distributed.BaguaModule]): Bagua distributed data parallel module.
        """

        # TODO: Remove the following code when BaguaModule is completely obsolete
        if type(bagua_ddp) is BaguaDistributedDataParallel:
            pass
        elif hasattr(bagua_ddp, "inner"):
            # Is bagua.torch_api.data_parallel.DistributedDataParallel
            bagua_ddp = bagua_ddp.inner
        elif hasattr(bagua_ddp, "bagua_ddp"):
            # Is bagua.torch_api.distributed.BaguaModule
            bagua_ddp = bagua_ddp.bagua_ddp
        else:
            raise Exception(
                "Unexpect input bagua_ddp({}), it should be BaguaDistributedDataParallel or bagua.torch_api.distributed.BaguaModule.".format(
                    type(bagua_ddp)
                )
            )

        if self.status in [_AsyncInternalState.NEW, _AsyncInternalState.STOPPED]:
            barrier(comm=self.process_group.get_global_communicator())
            if hasattr(bagua_ddp.bagua_buckets[0], "_async_op"):
                bagua_ddp.bagua_buckets[0]._async_op.reset()
            self.future = self.executor.submit(self._run_async_loop, bagua_ddp)
            self.status = _AsyncInternalState.SCHEDULED
            logging.debug("Process {} async communication resumed.".format(get_rank()))


class AsyncModelAverageAlgorithm(Algorithm):
    def __init__(
        self,
        peer_selection_mode: str = "all",
        sync_interval_ms: int = 500,
        warmup_steps: int = 0,
    ):
        """
        Create an instance of the
        `AsyncModelAverage <https://tutorials.baguasys.com/algorithms/async-model-average.html>`_
        algorithm.

        The asynchronous implementation is experimental, and imposes some restrictions.
        With such asynchronous algorithm, the number of iterations on each worker are different. Therefore
        the current implementation assumes that the dataset is an endless stream, and all workers continuously
        synchronize between each other.

        Users should call :meth:`abort` to manually stop the algorithm's continuous synchronization process.
        For example, for a model wrapped with `.with_bagua(...)`, you can abort with `model.bagua_algorithm.abort(model)`,
        and resume with `model.bagua_algorithm.resume(model)`.

        Args:
            peer_selection_mode (str): The way how workers communicate with each other. Currently ``"all"`` is supported.
                ``"all"`` means all workers' weights are synchronized during each communication.
            sync_interval_ms (int): Number of milliseconds between model synchronizations.
            warmup_steps (int): Number of steps to warm up by doing gradient allreduce before doing asynchronous
                model averaging. Use 0 to disable.
        """

        self.peer_selection_mode = peer_selection_mode
        self.sync_interval_ms = sync_interval_ms
        self.warmup_steps = warmup_steps

    def reify(self, process_group: BaguaProcessGroup) -> AsyncModelAverageAlgorithmImpl:
        return AsyncModelAverageAlgorithmImpl(
            process_group,
            peer_selection_mode=self.peer_selection_mode,
            sync_interval_ms=self.sync_interval_ms,
            warmup_steps=self.warmup_steps,
        )



from torch.utils.data.dataset import Dataset
from .cache_loader import CacheLoader

__all__ = ["CachedDataset"]


class CachedDataset(Dataset):
    def __init__(
        self,
        dataset: Dataset,
        backend: str = "redis",
        dataset_name: str = "",
        writer_buffer_size: int = 20,
        **kwargs,
    ):
        """
        Cached dataset wraps a `PyTorch dataset <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.Dataset>`_
        to cache its samples in memory, so that accessing these samples after the
        first time can be much faster. This is useful when samples need tedious preprocessing to produce, or reading
        the dataset itself is slow, which could slow down the whole training process.

        Internally, the samples are indexed by a string key ``"{dataset_name}_{index}"`` and saved in a distributed key-value
        store, where ``dataset_name`` is specified when initializing the cached dataset, and ``index`` is the index
        of a specific sample (the argument of :meth:`__getitem__` method in a PyTorch dataset).

        Args:
            dataset: PyTorch dataset to be wrapped.
            backend(str): Backend distributed key-value store implementation. Can be ``"redis"``.
            dataset_name(str): Name of the dataset. Default ``""``.
            writer_buffer_size(int): Number of samples to collect before writing to the backend key-value store.
                Useful for improving the backend throughput.

        Example::

            >>> from bagua.torch_api.contrib import CachedDataset
            >>> cache_dataset = CachedDataset(dataset, backend="redis", dataset_name="ds")
            >>> dataloader = torch.utils.data.DataLoader(cached_dataset)

        .. note::
            Cached dataset is a special case of cache loader. Parameter :attr:`backend` and :attr:`writer_buffer_size` in
            initializing a cached dataset have the same meanings as those in initializing a cache loader. You can
            provide the arguments for cache loader here in ``**kwargs``. See also :class:`~bagua.torch_api.contrib.cache_loader.CacheLoader`.

        """

        self.dataset = dataset

        self.cache_loader = CacheLoader(
            backend,
            dataset_name,
            writer_buffer_size,
            **kwargs,
        )
        """
        The backend cache loader instance.
        """

    def __getitem__(self, item):
        return self.cache_loader.get(item, lambda x: self.dataset[x])

    def __len__(self):
        return len(self.dataset)
